open HolKernel boolLib Parse bossLib stringTheory pred_setTheory pairTheory res_quanTheory PairRules optionTheory relationTheory quantHeuristicsLib pairSimps pred_setSimps proofManagerLib


open numTheory
open arithmeticTheory


open Defn
open listTheory

open set_relationTheory;
     
open c11Theory;
open c11_base_intermediateTheory;


open cpp11Theory;
open cpp11soundnessTheory;

open herdTheory;

open lem_relationTheory


val _ = new_theory "meta"

fun fs thms = FULL_SIMP_TAC (srw_ss()) thms;
fun gs thms = FULL_SIMP_TAC (std_ss ++ gen_beta_ss) thms;
fun stdfs thms = FULL_SIMP_TAC (std_ss) thms;
fun stds thms = SIMP_TAC (std_ss) thms;
fun ps thms = FULL_SIMP_TAC (std_ss ++ PRED_SET_ss) thms;
fun psg thms = SIMP_TAC (std_ss ++ PRED_SET_ss) thms;
fun rw thms = RW_TAC (srw_ss()) thms;
val st      = REPEAT STRIP_TAC;
val dg      = REPEAT (DISCH_TAC ORELSE GEN_TAC);
fun show1 term tactic = term by ALL_TAC THEN1 tactic;
fun ts thms = FULL_SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) thms;
fun simp_ts thms = SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) thms;
fun asm_ts thms = ASM_SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) thms;

fun mppat quote = Q.PAT_ASSUM quote MP_TAC;
fun rmpat quote = Q.PAT_ASSUM quote (fn x => ALL_TAC);
fun specpat quote specq = Q.PAT_ASSUM quote (fn x => ASSUME_TAC (Q.SPEC specq x));
fun specpatadd quote specq = Q.PAT_ASSUM quote
  (fn x => ASSUME_TAC (x) THEN ASSUME_TAC (Q.SPEC specq x));
fun mpspecpat quote specq = Q.PAT_ASSUM quote (fn x => MP_TAC (Q.SPEC specq x));

fun homatchmppat quote = Q.PAT_ASSUM quote HO_MATCH_MP_TAC;
fun matchmppat quote = Q.PAT_ASSUM quote MATCH_MP_TAC;

fun specpatp quote q1 q2 = Q.PAT_ASSUM quote
                           (fn x => ASSUME_TAC (Q.SPEC q2 (Q.SPEC q1 x)));
fun gs thms = SIMP_TAC (srw_ss()) thms;
fun met thms = METIS_TAC thms;
fun pats quote thms = Q.PAT_ASSUM quote (fn x => ASSUME_TAC (SIMP_RULE (srw_ss()) thms x));

fun impresp quote = Q.PAT_ASSUM quote IMP_RES_TAC;

fun popspec quote = POP_ASSUM (fn x => ASSUME_TAC (Q.SPEC quote x));

fun popsimp thms = POP_ASSUM (fn x => ASSUME_TAC (SIMP_RULE (srw_ss()) thms x));



val (apply_tree_rw, apply_tree_ind_rw) =
  Defn.tprove_no_defn ((apply_tree_def, apply_tree_ind),
  Q.EXISTS_TAC `measure (\ (t,_). named_predicate_tree_size t)` THEN
  fs [] THEN
  Induct_on `l` THENL
  [ fs [],
    FULL_SIMP_TAC list_ss [named_predicate_tree_size_def, DISJ_IMP_THM, FORALL_AND_THM] THEN
    REPEAT STRIP_TAC THEN
    RES_TAC THEN
    FULL_SIMP_TAC arith_ss []
  ]
);
val apply_tree_rw = save_thm ("apply_tree_rw",apply_tree_rw);
val apply_tree_ind_rw = save_thm ("apply_tree_ind_rw",apply_tree_ind_rw);




val swlem = Define `swlem Xlp Xlw = 
    let hrs   = hypothetical_release_sequence_set Xlp.actions Xlp.lk Xlw.mo in
    let rs    = release_sequence_set Xlp.actions Xlp.lk Xlw.mo in
    let sw    = release_acquire_fenced_synchronizes_with_set Xlp.actions Xlp.sb Xlp.asw Xlw.rf Xlw.lo rs hrs in
    sw`


val doblem = Define `doblem Xlp Xlw = 
    let rs    = release_sequence_set Xlp.actions Xlp.lk Xlw.mo in
    let cad   = with_consume_cad_set Xlp.actions Xlp.sb Xlp.dd Xlw.rf in
    let dob   = with_consume_dob_set Xlp.actions Xlw.rf rs cad in
    dob`



val hblem = Define `hblem Xlp Xlw = 
    let sw    = swlem Xlp Xlw in
    let dob   = doblem Xlp Xlw in
    let ithb  = inter_thread_happens_before Xlp.actions Xlp.sb sw dob in
    let hb    = happens_before Xlp.actions Xlp.sb ithb in
    hb`




val relacq_writeread_atomic = store_thm("relacq_writeread_atomic",
``(!a. (is_write a /\ is_release a) \/ (is_read a /\ (is_acquire a \/ is_consume a)) ==> is_atomic_action a)``,
  rw [] THEN
  ps [is_release_def,is_write_def,is_acquire_def,is_consume_def,is_read_def,is_atomic_action_def] THEN
  Cases_on `a` THEN
  fs []
)



val scWriteIsRelease = store_thm("scWriteIsRelease",
``(!a. is_write a /\ is_seq_cst a ==> is_release a)``,
  rw [] THEN
  ps [is_release_def,is_write_def,is_seq_cst_def] THEN
  Cases_on `a` THEN
  fs []
)


val scReadIsAcquire = store_thm("scReadIsAcquire",
``(!a. is_read a /\ is_seq_cst a ==> is_acquire a)``,
  rw [] THEN
  ps [is_acquire_def,is_read_def,is_seq_cst_def] THEN
  Cases_on `a` THEN
  fs []
)



val nonatomic_not_atomic = store_thm("nonatomic_not_atomic",
``(!a. is_non_atomic_action a ==> ~is_atomic_action a)``,
  rw [] THEN
  ps [is_non_atomic_action_def,is_NA_load_def,is_NA_store_def,is_atomic_action_def] THEN
  Cases_on `a` THEN
  fs []
)





val qsimps =
  [ RES_FORALL,
    RES_EXISTS,
    FORALL_PROD,
    EXISTS_PROD
  ]


val cacq_rw = store_thm("cacq_rw",
``!X a. cacq X a = is_acquire a /\ a IN X.acts``,
  rw [] THEN
  ps [cacq_def,acq_def,sc_def,acq_rel_def,F_def,con_def,is_acquire_def,is_herd_acquire_def,is_seq_cst_def,is_herd_acq_rel_def,is_consume_def,is_fence_def,R_def,is_read_def,is_herd_consume_def] THEN
  Cases_on `a` THEN
  fs [] THEN
  met []
)


val cacq_rw2 = store_thm("cacq_rw",
``!X a. cacq X = {a | is_acquire a /\ a IN X.acts}``,
  rw [] THEN
  ps [cacq_def,acq_def,sc_def,acq_rel_def,F_def,con_def,is_acquire_def,is_herd_acquire_def,is_seq_cst_def,is_herd_acq_rel_def,is_consume_def,is_fence_def,R_def,is_read_def,is_herd_consume_def,EXTENSION] THEN
  rw [] THEN
  Cases_on `x` THEN
  fs [] THEN
  met []
)




val crel_rw = store_thm("crel_rw",
``!X a. crel X a = is_release a /\ a IN X.acts``,
  rw [] THEN
  ps [crel_def,rel_def,sc_def,acq_rel_def,F_def,con_def,is_release_def,is_herd_release_def,is_seq_cst_def,is_herd_acq_rel_def,is_consume_def,is_fence_def,W_def,is_write_def] THEN
  Cases_on `a` THEN
  fs [] THEN
  met []
)



val crel_rw2 = store_thm("crel_rw",
``!X a. crel X = {a | is_release a /\ a IN X.acts}``,
  rw [] THEN
  ps [crel_def,rel_def,sc_def,acq_rel_def,F_def,con_def,is_release_def,is_herd_release_def,is_seq_cst_def,is_herd_acq_rel_def,is_consume_def,is_fence_def,W_def,is_write_def,EXTENSION] THEN
  Cases_on `x` THEN
  fs [] THEN
  met []
)


val ccon_rw = store_thm("ccon_rw",
``!X a. ccon X a = is_consume a /\ a IN X.acts``,
  rw [] THEN
  ps [ccon_def,con_def,sc_def,acq_rel_def,F_def,con_def,is_consume_def,R_def,is_herd_consume_def,is_read_def] THEN
  Cases_on `a` THEN
  fs [] THEN
  met []
)


val ccon_rw2 = store_thm("ccon_rw",
``!X a. ccon X = {a | is_consume a /\ a IN X.acts}``,
  rw [] THEN
  ps [ccon_def,con_def,sc_def,acq_rel_def,F_def,con_def,is_consume_def,R_def,is_herd_consume_def,is_read_def,EXTENSION] THEN
  Cases_on `x` THEN
  fs [] THEN
  met []
)


val herdsimps =
  [ rmwv_def,corr_def,cowr_def,corw_def,coww_def,
    hbarf_def,arf_def,dd_def,
    unis_def,
    fr_def,
    narf_def,vse_def,hbl_def,loc_def,
    atomicloc_def,nonatomicloc_def,mutexloc_def,
    mv_def,inv_def,lem_converse_def,SWAP_def,ls_def,
    sc_def,
    sw_def,
    dob_def,
    cad_def,
    seq_def,
    swul_def,
    lk_def,
    ul_def,
    ext_def,
    pp_asw_def,
    compr_def,
    comps_def,
    sw_def,
    swra_def,
    fid_def,
    idf_def,
    rs_def,
    con_def,
    coe_def,
    coi_def,
    comps_def,
    compr_def,
    na_def,
    F_def,
    I_def,
    WW_def,
    WR_def,
    W_def,
    R_def,
    rc_def,
    rmw_def,
    int_def,
    ext_def,
    INTER_DEF,
    UNION_DEF,
    cross_def,
    rcomp_def,
    SUBSET_DEF,
    tch_def,
    stor_def,
    A_def,
    scco_def,
    scfr_def,
    schb_def,
    scp_def,
    herdTheory.irreflexive_def,
    lem_is_irreflexive_def,
    herdTheory.acyclic_def,
    tch_def,
    cacq_rw2,
    crel_rw2,
    ccon_rw2
  ] @ qsimps



val lemsimps =
  [ release_acquire_fenced_synchronizes_with_set_def,
    release_acquire_fenced_synchronizes_with_def,
    with_consume_dob_set_def,
    with_consume_cad_set_def,
    dependency_ordered_before_def,
    release_sequence_set_def,
    hypothetical_release_sequence_set_def,
    rs_element_def
  ] @ qsimps


val mysimps = lemsimps @ herdsimps



val sameLocThenAtomicLoc = store_thm("sameLocThenAtomicLoc",
``(
!a b Xo. (loc_of a = loc_of b) /\ is_at_atomic_location Xo.lk a ==>
      is_at_atomic_location Xo.lk b
)``,
  rw [] THEN
  ps [is_at_atomic_location_def] THEN
  met []
)




val lem_consistent_then_well_formed_actions = store_thm("lem_consistent_then_well_formed_actions",
``(
!Xlp Xlw rl.
  apply_tree with_consume_consistent_execution (Xlp,Xlw,rl)
  ==>
  (well_formed_actions Xlp.actions)
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def,with_consume_relations_def] THEN
  fs [apply_tree_rw,well_formed_threads_def] THEN
  ps [well_formed_actions_def,RES_FORALL]
)




val lem_consistent_then_partial_eq = store_thm("lem_consistent_then_partial_eq",
``(
!Xlp Xlw p herd_opsem opsem r rl addr data.
  let Xh = lemToHerdExecution Xlp Xlw addr data in
  opsem p Xlp /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xlp.dd) /\
  lemHerdCondition herd_opsem opsem p /\
  apply_tree with_consume_consistent_execution (Xlp,Xlw,("hb",r)::rl)
  ==>
  lemHerdPartialExecutionEquivalent (Xlp,Xlw) Xh
)``,
  rw [] THEN
  Q.UNABBREV_TAC `Xh` THEN
  psg [tch_def,pp_asw_def,dd_def,lemHerdPartialExecutionEquivalent_def,lemHerdPreExecutionEquivalent_def,lemToHerdExecution_def,compr_def,rcomp_def,seq_def,EXTENSION] THEN
  fs [LET_DEF,FORALL_PROD,UNION_DEF,EXTENSION] THEN
  Q.ABBREV_TAC `Xh = lemToHerdExecution Xlp Xlw addr data` THEN
  `well_formed_inits Xh.inits Xh.acts Xh.po` by met [lemHerdCondition_def] THEN
  ps [well_formed_inits_def,RES_FORALL,RES_EXISTS] THEN
  Q.UNABBREV_TAC `Xh` THEN
  fs [lemToHerdExecution_def,LET_DEF,SUBSET_DEF] THEN
  rw [] THEN
  met [] 
)



val differentTidsDifferentActions = store_thm("differentTidsDifferentActions",
``(
!a b. (tid_of a <> tid_of b) ==> (a <> b)
)``,
  rw [] THEN
  CCONTR_TAC THEN
  ps [] THEN
  rw []
)  








val NALocThenReadOrWrite = store_thm("NALocThenReadOrWrite",
``(
!Xo Xw rl a. apply_tree with_consume_consistent_execution (Xo,Xw,rl) /\
       is_at_non_atomic_location Xo.lk a /\ a IN Xo.actions ==>
       is_write a \/ is_read a
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,well_formed_threads_def,actions_respect_location_kinds_def,is_read_def,is_write_def,RES_FORALL,is_at_non_atomic_location_def,loc_of_def] THEN
  specpat `!a. a IN Xo.actions ==> Q` `a` THEN 
  Cases_on `a` THEN
  fs []
)





val sameLocThenNALoc = store_thm("sameLocThenNALoc",
``(
!a b Xo. (loc_of a = loc_of b) /\ is_at_non_atomic_location Xo.lk a ==>
      is_at_non_atomic_location Xo.lk b
)``,
  rw [] THEN
  ps [is_at_non_atomic_location_def] THEN
  met []
)






val readIsLoadOrRmw = store_thm("readIsLoadOrRmw",
``(
  !a. is_read a ==> is_load a \/ is_RMW a
)``,
  rw [] THEN
  ps [is_read_def,is_load_def,is_RMW_def] THEN
  Cases_on `a` THEN
  fs []
)

val reloversimps =
  [ rel_over_def,RES_FORALL,relation_over_def,range_def,domain_def,SUBSET_DEF ]


val relation_over_thm = Q.store_thm ("relation_over_thm",
`!s rel. relation_over s rel = !a b. (a,b) IN rel ==> a IN s /\ b IN s`,
  fs reloversimps THEN
  met []
)



val ithb_over_actions0 = store_thm("ithb_over_actions0",
``(!actions threads lk sbi asw dd cd sw dob.
    relation_over actions sbi /\ relation_over actions sw /\ relation_over actions dob ==>
    relation_over actions (inter_thread_happens_before actions sbi sw dob)
)``,
  rw [] THEN
  fs [inter_thread_happens_before_def, relation_over_thm, RES_FORALL, RES_EXISTS] THEN
  STRIP_TAC THEN
  STRIP_TAC THEN
  STRIP_TAC THEN
  `let r = sw' UNION dob' UNION (compose sw' sbi) in
     (a IN domain (r UNION (compose sbi r))) /\ (b IN range (r UNION (compose sbi r)))` by ALL_TAC THEN1
  ( fs [LET_DEF] THEN
    MATCH_MP_TAC tc_domain_range THEN
    met []
  ) THEN
  fs [LET_DEF,domain_def,range_def,compose_def] THEN
  met []
)


val ithb_over_actions = Q.store_thm ("ithb_over_actions",
`! Xo lo rs hrs rf cad.
  let sw  = release_acquire_fenced_synchronizes_with_set Xo.actions Xo.sb Xo.asw rf lo rs hrs in
  let dob = with_consume_dob_set Xo.actions rf rs cad in
    well_formed_threads (Xo,_,_) ==>
    relation_over Xo.actions (inter_thread_happens_before Xo.actions Xo.sb sw dob)`,
  rw [] THEN
  HO_MATCH_MP_TAC ithb_over_actions0 THEN
  gs [RES_FORALL,RES_EXISTS,LET_DEF,relation_over_thm] THEN
  fs [well_formed_threads_def,RES_FORALL,RES_EXISTS,LET_DEF,threadwise_def,relation_over_thm,FORALL_PROD] THEN
  fs [release_acquire_fenced_synchronizes_with_set_def,release_acquire_fenced_synchronizes_with_def,
        with_consume_dob_set_def,with_consume_dob_def] THEN
Q.UNABBREV_TAC `sw'` THEN
Q.UNABBREV_TAC `dob'` THEN
fs [EXTENSION] THEN
met [])



val hb_over_actions = Q.store_thm ("hb_over_actions",
`! Xo ithb'.
  let hb = happens_before Xo.actions Xo.sb ithb' in
    well_formed_threads (Xo,_,_) /\ relation_over Xo.actions ithb' ==>
    relation_over Xo.actions hb`,
  rw [] THEN
  `relation_over Xo.actions Xo.sb` by fs [relation_over_thm,well_formed_threads_def,FORALL_PROD,RES_FORALL] THEN1 met [] THEN
  Q.UNABBREV_TAC `hb'` THEN
  fs [happens_before_def,relation_over_thm,RES_FORALL] THEN
  rw [] THEN
  FULL_SIMP_TAC (std_ss ++ pairSimps.gen_beta_ss) [] THEN
  met []
)







val inv_thm = Q.store_thm ("inv_thm",
`
!a b r. (a,b) IN inv r
      ==> (b,a) IN r
`,
  rw [] THEN
  ps [inv_def,lem_converse_def,SWAP_def]
)



val SCAtAtomicLoc = Q.store_thm ("SCAtAtomicLoc",
`
!actions lk a.
actions_respect_location_kinds actions lk /\ (is_write a \/ is_read a) /\ is_seq_cst a /\ a IN actions ==>
is_at_atomic_location lk a`,
  rw [] THEN
  ps [actions_respect_location_kinds_def,RES_FORALL] THEN
  specpat `!a. Q` `a` THEN mppat `Q` THEN ps [] THEN rw [] THEN
  ps [is_seq_cst_def,is_at_atomic_location_def,RES_FORALL,loc_of_def,is_write_def,is_read_def] THEN
  Cases_on `a` THEN
  fs [] THEN
  rw []
)




val rmwAtAtomicLoc = Q.store_thm ("rmwAtAtomicLoc",
`
!actions lk a.
actions_respect_location_kinds actions lk /\ is_RMW a /\ a IN actions ==>
is_at_atomic_location lk a`,
  rw [] THEN
  ps [actions_respect_location_kinds_def,RES_FORALL] THEN
  specpat `!a. Q` `a` THEN mppat `Q` THEN ps [] THEN rw [] THEN
  ps [is_at_atomic_location_def,RES_FORALL,loc_of_def,is_RMW_def] THEN
  Cases_on `a` THEN
  fs [] THEN
  rw []
)




val readWriteIsRmw = Q.store_thm ("readWriteIsRmw",
`
!a. is_read a /\ is_write a ==> is_RMW a`,
  rw [] THEN
  ps [is_read_def,is_write_def,is_RMW_def] THEN
  Cases_on `a` THEN
  fs []
)










val lem_consistent_then_hb_requirements = store_thm("lem_consistent_then_hb_requirements",
``(
!Xlp Xlw p herd_opsem opsem r rl addr data.
  let Xh = lemToHerdExecution Xlp Xlw addr data in
  opsem p Xlp /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xlp.dd) /\
  lemHerdCondition herd_opsem opsem p /\
  apply_tree with_consume_consistent_execution (Xlp,Xlw,("hb",r)::rl)
  ==>
  (well_formed_actions Xh.acts) /\
  (!a b. (a,b) IN pp_asw Xh ==> (tid_of a <> tid_of b)) /\
  (!a b. a IN Xh.inits /\ b IN Xh.acts /\ b NOTIN Xh.inits ==> (tid_of a <> tid_of b)) /\
  Xh.rfh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.loh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.po  SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (cross Xh.acts Xh.acts) /\
  dd Xh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (WW Xh) /\
  Xh.rfh SUBSET (WR Xh) /\
  Xh.aswh  SUBSET (cross Xh.acts Xh.acts)
)``,
  fs [LET_DEF,lemToHerdExecution_def,WR_def,R_def,W_def,rel_over_def,cross_def,well_formed_rf_def,RES_FORALL,SUBSET_DEF,relation_over_def,domain_def,range_def,FORALL_PROD,dd_def,WW_def,pp_asw_def,compr_def,rcomp_def,seq_def,well_formed_actions_def,tch_def] THEN
  rw [] THEN
  Q.ABBREV_TAC `Xh = <|inits := {a | ?b. (a,b) IN Xlp.asw}; acts := Xlp.actions;
          lkh := Xlp.lk; po := Xlp.sb;
          aswh :=
            {(a,b) |
             (?b. (a,b) IN Xlp.asw) /\ (!b'. (b,b') NOTIN Xlp.asw) /\
             b IN Xlp.actions}; addr := addr; data := data;
          rfh := Xlw.rf; co := Xlw.mo; loh := Xlw.lo|>` THEN
  `well_formed_inits Xh.inits Xh.acts Xh.po` by met [lemHerdCondition_def] THEN
  Q.UNABBREV_TAC `Xh` THEN
  fs [well_formed_inits_def,RES_FORALL,RES_EXISTS,SUBSET_DEF] THEN
  ps [with_consume_consistent_execution_def,with_consume_relations_def] THEN
  fs [apply_tree_rw,well_formed_threads_def,RES_FORALL,FORALL_PROD] THEN
  fs [consistent_mo_def,relation_over_def,domain_def,range_def,rel_over_def,locks_only_consistent_lo_def,well_formed_rf_def,RES_FORALL,SUBSET_DEF,FORALL_PROD] THEN
  met [IN_UNION]
)





val swEquivalent = store_thm("swEquivalent",
``(
! Xh Xlp Xlw.
  (well_formed_actions Xh.acts) /\
  (!a b. (a,b) IN pp_asw Xh ==> (tid_of a <> tid_of b)) /\
  (!a b. a IN Xh.inits /\ b IN Xh.acts /\ b NOTIN Xh.inits ==> (tid_of a <> tid_of b)) /\
  Xh.loh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.po  SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (cross Xh.acts Xh.acts)/\
  Xh.aswh  SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (WW Xh) /\
  Xh.rfh SUBSET (WR Xh) /\
  lemHerdPartialExecutionEquivalent (Xlp,Xlw) Xh
  ==>
  (swlem Xlp Xlw = sw Xh)
)``,
  rw [] THEN
  ps [lemHerdPartialExecutionEquivalent_def,lemHerdPreExecutionEquivalent_def,hblem] THEN rw [] THEN
  psg [EXTENSION,FORALL_PROD] THEN rw [] THEN
  ps [swlem] THEN rw [] THEN
  Q.UNABBREV_TAC `sw'` THEN
  EQ_TAC THEN1
  ( rw [] THEN
    ps mysimps THEN
    `p_1 IN Xh.acts /\ p_2 IN Xh.acts` by met [] THEN
    Q.UNABBREV_TAC `hrs` THEN Q.UNABBREV_TAC `rs'` THEN ps [] THEN rw [] THEN
    DISJ2_TAC THEN DISJ2_TAC THENL
    [ met [A_def,relacq_writeread_atomic],
      ( Q.EXISTS_TAC `c` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      ),
      ( Q.EXISTS_TAC `c` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      ),
      ( Q.EXISTS_TAC `x` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      ),
      ( Q.EXISTS_TAC `z` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      ),
      ( Q.EXISTS_TAC `z` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      ),
      met [A_def,relacq_writeread_atomic],
      ( Q.EXISTS_TAC `y` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      ),
      ( Q.EXISTS_TAC `y` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      ),
      met [A_def,relacq_writeread_atomic],
      ( Q.EXISTS_TAC `y` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      ),
      ( Q.EXISTS_TAC `y` THEN rw [] THEN
        met [A_def,relacq_writeread_atomic]
      )
    ]
  ) THEN
  rw [] THEN
  ps mysimps THEN
  `p_1 IN Xh.acts /\ p_2 IN Xh.acts` by met [] THEN
  Q.UNABBREV_TAC `hrs` THEN Q.UNABBREV_TAC `rs'` THEN ps [] THEN rw [] THENL
  [ ( DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ1_TAC THEN
      Q.EXISTS_TAC `z'` THEN rw [] THEN
      met [A_def,relacq_writeread_atomic]
    ),
    ( DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ1_TAC THEN
      Q.EXISTS_TAC `z'` THEN rw [] THEN
      met [A_def,relacq_writeread_atomic]
    ),
    ( DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ1_TAC THEN
      Q.EXISTS_TAC `z'` THEN rw [] THEN
      met [A_def,relacq_writeread_atomic]
    ),
    ( DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ1_TAC THEN
      Q.EXISTS_TAC `z'` THEN rw [] THEN
      met [A_def,relacq_writeread_atomic]
    ),
    met [A_def,relacq_writeread_atomic],
    met [A_def,relacq_writeread_atomic],
    met [A_def,relacq_writeread_atomic],
    met [A_def,relacq_writeread_atomic],
    met [A_def,relacq_writeread_atomic],
    met [A_def,relacq_writeread_atomic],
    met [A_def,relacq_writeread_atomic],
    met [A_def,relacq_writeread_atomic]
  ]
)




val tcrfpodd_in_actions = store_thm("tcrfpodd_in_actions",
``(
! Xh a b.
  (!p_1' p_2'. (p_1',p_2') IN Xh.po ==> p_1' IN Xh.acts /\ p_2' IN Xh.acts) /\
  (!p_1' p_2'. (p_1',p_2') IN dd Xh ==> p_1' IN Xh.acts /\ p_2' IN Xh.acts) /\
  (a,b) IN tc {x | x IN Xh.rfh /\ x IN Xh.po \/ x IN dd Xh}
  ==>
  b IN Xh.acts
)``,
  rw [] THEN
  `?c. (c,b) IN {x | x IN Xh.rfh /\ x IN Xh.po \/ x IN dd Xh}` by met [tc_cases_right] THEN
  ps [EXTENSION,FORALL_PROD] THEN
  met []
)

val write_not_fence = store_thm("write_not_fence",
``(
!a. is_write a ==> ~is_fence a
)``,
  rw [] THEN
  ps [is_write_def,is_fence_def] THEN
  Cases_on `a` THEN
  fs []
)




val dobEquivalent = store_thm("dobEquivalent",
``(
! Xh Xlp Xlw.
  (well_formed_actions Xh.acts) /\
  (!a b. a IN Xh.inits /\ b IN Xh.acts /\ b NOTIN Xh.inits ==> (tid_of a <> tid_of b)) /\
  Xh.rfh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.po  SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (cross Xh.acts Xh.acts) /\
  dd Xh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (WW Xh) /\
  Xh.rfh SUBSET (WR Xh) /\
  lemHerdPartialExecutionEquivalent (Xlp,Xlw) Xh
  ==>
  (doblem Xlp Xlw = dob Xh)
)``,
  rw [] THEN
  `!j k. (j,k) IN tc
         {x |
          x IN Xh.rfh /\ x IN Xh.po \/
          x IN tc {x | x IN Xh.addr \/ x IN Xh.data}} ==>
   ?k'. (k',k) IN          {x |
          x IN Xh.rfh /\ x IN Xh.po \/
          x IN tc {x | x IN Xh.addr \/ x IN Xh.data}}` by met [tc_cases_right] THEN
  ps [lemHerdPartialExecutionEquivalent_def,lemHerdPreExecutionEquivalent_def,hblem] THEN rw [] THEN
  psg [EXTENSION,FORALL_PROD] THEN rw [] THEN
  ps [doblem] THEN rw [] THEN
  Q.UNABBREV_TAC `dob'` THEN
  Q.UNABBREV_TAC `cad'` THEN
  EQ_TAC THEN1
  ( rw [] THEN
    ps mysimps THEN
    Q.UNABBREV_TAC `rs'` THEN ps [] THEN rw [] THEN
    Q.EXISTS_TAC `b'` THEN rw [] THEN
    met [A_def,relacq_writeread_atomic]
  ) THEN
  rw [] THEN
  ps mysimps THEN
  Q.UNABBREV_TAC `rs'` THEN ps [] THEN rw [] THEN
  `p_1 IN Xh.acts` by met [] THEN
  `p_2 IN Xh.acts` by
  ( `p_2 IN Xh.acts \/ ?d. (d,p_2) IN {x |
          x IN Xh.rfh /\ x IN Xh.po \/
          x IN tc {x | x IN Xh.data \/ x IN Xh.addr}}
    ` by met [tc_cases_right] THEN
    fs [] THEN
    met [tc_cases_left]
  ) THEN
  met [A_def,write_not_fence,relacq_writeread_atomic]
)




val seqComposeEq = store_thm("seqComposeEq",
``(! r1 r2. seq r1 r2 = compose r1 r2)``,
  rw [] THEN
  ps [seq_def,compose_def,rcomp_def,EXTENSION] THEN
  rw [] THEN
  met []
)




val hbEquivalent = store_thm("hbEquivalent",
``(
! Xh Xlp Xlw.
  (well_formed_actions Xh.acts) /\
  (!a b. (a,b) IN pp_asw Xh ==> (tid_of a <> tid_of b)) /\
  (!a b. a IN Xh.inits /\ b IN Xh.acts /\ b NOTIN Xh.inits ==> (tid_of a <> tid_of b)) /\
  Xh.rfh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.loh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.po  SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (cross Xh.acts Xh.acts) /\
  Xh.aswh  SUBSET (cross Xh.acts Xh.acts) /\
  dd Xh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (WW Xh) /\
  Xh.rfh SUBSET (WR Xh) /\
  lemHerdPartialExecutionEquivalent (Xlp,Xlw) Xh
  ==>
  (hblem Xlp Xlw = hb Xh)
)``,
  rw [] THEN
  `swlem Xlp Xlw = sw Xh` by met [swEquivalent] THEN
  `doblem Xlp Xlw = dob Xh` by met [dobEquivalent] THEN
  ps [lemHerdPartialExecutionEquivalent_def,lemHerdPreExecutionEquivalent_def,hblem] THEN rw [] THEN
  psg [EXTENSION,FORALL_PROD] THEN rw [] THEN
  Q.UNABBREV_TAC `hb'` THEN
  Q.UNABBREV_TAC `ithb'` THEN
  ps [happens_before_def,inter_thread_happens_before_def,hb_def,ithb_def,ithbr_def,LET_DEF,tch_def] THEN
  rw [seqComposeEq] THEN
  met [UNION_ASSOC]
)











val lem_consistent_then_hb_eq = store_thm("lem_consistent_then_hb_eq",
``(
!Xlp Xlw p herd_opsem opsem r rl addr data.
  let Xh = lemToHerdExecution Xlp Xlw addr data in
  opsem p Xlp /\
  herd_opsem p Xh /\
  (tc(data UNION addr) = Xlp.dd) /\
  lemHerdCondition herd_opsem opsem p /\
  apply_tree with_consume_consistent_execution (Xlp,Xlw,("hb",r)::rl)
  ==>
  (hblem Xlp Xlw = hb Xh)
)``,
  rw [] THEN
  `lemHerdPartialExecutionEquivalent (Xlp,Xlw) Xh` by met [lem_consistent_then_partial_eq,LET_DEF] THEN
  ASSUME_TAC lem_consistent_then_hb_requirements THEN
  fs [LET_DEF] THEN
  RES_TAC THEN
  met [hbEquivalent]
)



val lemHerdExistsEquivalentExecution = store_thm("lemHerdExistsEquivalentExecution",
``(
!Xo Xw opsem herd_opsem p.
lemHerdCondition herd_opsem opsem p /\
opsem p Xo /\
apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw)
==>
?data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  herd_opsem p Xh /\
  (tc(data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
)``,
  rw [] THEN
  `?data addr. let Xh = lemToHerdExecution Xo Xw addr data in
    herd_opsem p Xh /\
    (tc (data UNION addr) = Xo.dd)` by met [lemHerdCondition_def,tch_def] THEN
  Q.EXISTS_TAC `data` THEN
  Q.EXISTS_TAC `addr` THEN
  ps [LET_DEF] THEN
  STRIP_TAC THEN
  STRIP_TAC THEN
  `hblem Xo Xw = hb Xh` by ALL_TAC THEN1
  ( fs [with_consume_relations_def] THEN
    ASSUME_TAC lem_consistent_then_hb_eq THEN
    fs [LET_DEF] THEN
    met []
  ) THEN
  ps [scp_def,SUBSET_DEF,FORALL_PROD] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw] THEN
  rw [] THENL
  [ ( ps herdsimps THEN
      pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
      pats `consistent_hb Q` [consistent_hb_def,RES_FORALL,with_consume_relations_def,LET_DEF,lem_is_irreflexive_def,FORALL_PROD] THEN
      ps [hblem,swlem,doblem,LET_DEF] THEN
      `p_1 <> p_2` by met [tc_rules] THEN
      `p_1 IN Xo.actions /\ p_2 IN Xo.actions` by ALL_TAC THEN1
      ( `relation_over Xo.actions (hb Xh)` by met [hb_over_actions,ithb_over_actions,LET_DEF] THEN
        ps reloversimps THEN
        met []
      ) THEN
      met []
    ),
    ( ps herdsimps THEN
      pats `consistent_mo Q` [consistent_mo_def,RES_FORALL] THEN
      ps [hblem,swlem,doblem,LET_DEF] THEN
      ps reloversimps THEN
      pats `well_formed_threads Q` [well_formed_threads_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
      ps [] THEN
      pats `relation_over Xo.actions Xo.sb` [relation_over_def,rel_over_def,domain_def,range_def,SUBSET_DEF] THEN
      pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
      pats `coherent_memory_use Q` [coherent_memory_use_def,RES_FORALL,with_consume_relations_def,LET_DEF,RES_EXISTS,EXISTS_PROD] THEN
      pats `lem_is_strict_partial_order Xo.sb` [lem_is_strict_partial_order_def,transitive_def] THEN
      Q.UNABBREV_TAC `Xh` THEN
      fs [lemToHerdExecution_def,LET_DEF] THEN
      `p_1 <> p_2` by met [happens_before_def,IN_UNION] THEN
      `(p_1,p_2) IN Xw.sc \/ (p_2,p_1) IN Xw.sc` by met [] THEN
      `p_1 IN Xo.actions /\ p_2 IN Xo.actions` by met [] THEN
      pats `sc_fenced_sc_fences_heeded Q` [sc_fenced_sc_fences_heeded_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
      mppat `Q` THEN ps [] THEN rw [] THEN
      met []
      (* This takes a long time *)
    ),
    ( fs herdsimps THEN
      ps [hblem,swlem,doblem,LET_DEF] THEN
      Q.ABBREV_TAC `hrs =
                  hypothetical_release_sequence_set Xo.actions Xo.lk
                    Xw.mo` THEN
      Q.ABBREV_TAC `rs' = release_sequence_set Xo.actions Xo.lk Xw.mo` THEN
      Q.ABBREV_TAC `sw' =
                  release_acquire_fenced_synchronizes_with_set
                    Xo.actions Xo.sb Xo.asw Xw.rf Xw.lo rs' hrs
            ` THEN
      Q.ABBREV_TAC `cad' = with_consume_cad_set Xo.actions Xo.sb Xo.dd Xw.rf` THEN
      Q.ABBREV_TAC `dob' = with_consume_dob_set Xo.actions Xw.rf rs' cad'` THEN
      Q.ABBREV_TAC `ithb' =
                  inter_thread_happens_before Xo.actions Xo.sb sw' dob'` THEN
      Q.ABBREV_TAC `hb' = happens_before Xo.actions Xo.sb ithb'` THEN
      Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
      ps [with_consume_consistent_execution_def] THEN
      fs [apply_tree_rw] THEN
      Q.ABBREV_TAC `hb'' = hb (lemToHerdExecution Xo Xw addr data)` THEN
      `relation_over Xo.actions (hb Xh)` by met [hb_over_actions,ithb_over_actions,LET_DEF] THEN
      Q.UNABBREV_TAC `Xh` THEN
      fs [lemToHerdExecution_def,LET_DEF] THENL
      [ ( `is_write z /\ is_read p_1 /\ z IN Xo.actions /\ p_1 IN Xo.actions /\ (loc_of z = loc_of p_1)` by ALL_TAC THEN1
          ( pats `well_formed_rf Q` [well_formed_rf_def,RES_FORALL,FORALL_PROD] THEN
            ps [relation_over_def,rel_over_def,domain_def,range_def,SUBSET_DEF] THEN
            met []
          ) THEN
          `is_at_atomic_location Xo.lk p_1` by ALL_TAC THEN1
          ( pats `well_formed_threads Q` [well_formed_threads_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
            ps [] THEN
            met [SCAtAtomicLoc]
          ) THEN
          `is_at_atomic_location Xo.lk p_2` by met [sameLocThenAtomicLoc] THEN
          `z <> p_2` by ALL_TAC THEN1
          ( pats `consistent_hb Q` [consistent_hb_def,RES_FORALL,with_consume_relations_def,LET_DEF,lem_is_irreflexive_def,FORALL_PROD] THEN
            met [tc_cases]
          ) THEN
          `p_2 IN Xo.actions` by ALL_TAC THEN1
          ( ps reloversimps THEN
            met []
          ) THEN
          `(p_2,z) IN Xw.mo \/ (z,p_2) IN Xw.mo` by ALL_TAC THEN1
          ( pats `consistent_mo Q` [consistent_mo_def,RES_FORALL] THEN
            met []
          ) THEN1
          ( pats `coherent_memory_use Q` [coherent_memory_use_def,RES_FORALL,with_consume_relations_def,LET_DEF,RES_EXISTS,EXISTS_PROD] THEN
            met []
          ) THEN
          `(p_1,p_2) IN Xw.sc \/ (p_2,p_1) IN Xw.sc` by ALL_TAC THEN1
          ( pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
            met []
          ) THEN
          pats `sc_accesses_sc_reads_restricted Q` [sc_accesses_sc_reads_restricted_def,RES_FORALL,with_consume_relations_def,LET_DEF,FORALL_PROD,RES_EXISTS,EXISTS_PROD] THEN
          specpat `Q` `z` THEN
          specpat `Q` `p_1` THEN
          mppat `Q` THEN ps [] THEN rw [] THEN1
          ( specpat `Q` `p_2` THEN
            rw [] THEN
            pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
            met []
          ) THEN
          met []
        ),
        ( `is_write z /\ is_read p_1 /\ z IN Xo.actions /\ p_1 IN Xo.actions /\ (loc_of z = loc_of p_1)` by ALL_TAC THEN1
          ( pats `well_formed_rf Q` [well_formed_rf_def,RES_FORALL,FORALL_PROD] THEN
            ps reloversimps THEN
            met []
          ) THEN
          `is_at_atomic_location Xo.lk p_2 /\ (p_2 <> z) /\ (loc_of p_2 = loc_of z) /\ is_write p_2 /\ is_write z /\ p_2 IN Xo.actions /\ z IN Xo.actions` by ALL_TAC THEN1
          ( pats `consistent_mo Q` [consistent_mo_def,RES_FORALL,relation_over_def,rel_over_def,domain_def,range_def,SUBSET_DEF] THEN
            met []
          ) THEN
          `is_at_atomic_location Xo.lk p_1` by met [sameLocThenAtomicLoc] THEN
          `(p_1,p_2) IN Xw.sc \/ (p_2,p_1) IN Xw.sc` by ALL_TAC THEN1
          ( pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
            met []
          ) THEN
          pats `sc_accesses_sc_reads_restricted Q` [sc_accesses_sc_reads_restricted_def,RES_FORALL,with_consume_relations_def,LET_DEF,FORALL_PROD,RES_EXISTS,EXISTS_PROD] THEN
          specpat `Q` `z` THEN
          specpat `Q` `p_1` THEN
          mppat `Q` THEN ps [] THEN rw [] THEN
          specpat `Q` `p_2` THEN
          rw [] THEN
          pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
           met []
        ),
        ( `(p_1,p_2) IN Xw.sc \/ (p_2,p_1) IN Xw.sc` by ALL_TAC THEN1
          ( pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
            met []
          ) THEN
          `p_1 IN Xo.actions /\ p_2 IN Xo.actions /\ z IN Xo.actions /\ z' IN Xo.actions` by ALL_TAC THEN1
          ( pats `well_formed_rf Q` [well_formed_rf_def] THEN
            pats `consistent_mo Q` [consistent_mo_def] THEN
            pats `well_formed_threads Q` [well_formed_threads_def,RES_FORALL] THEN
            ps [relation_over_def,rel_over_def,domain_def,range_def,SUBSET_DEF,RES_FORALL,FORALL_PROD] THEN
            met []
          ) THEN
          pats `sc_fenced_sc_fences_heeded Q` [sc_fenced_sc_fences_heeded_def,RES_FORALL,RES_EXISTS] THEN          
          met []
        ),
        ( `(p_1,p_2) IN Xw.sc \/ (p_2,p_1) IN Xw.sc` by ALL_TAC THEN1
          ( pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
            met []
          ) THEN
          `p_1 IN Xo.actions /\ p_2 IN Xo.actions /\ z IN Xo.actions /\ z' IN Xo.actions` by ALL_TAC THEN1
          ( pats `well_formed_rf Q` [well_formed_rf_def] THEN
            pats `consistent_mo Q` [consistent_mo_def] THEN
            pats `well_formed_threads Q` [well_formed_threads_def,RES_FORALL] THEN
            ps [relation_over_def,rel_over_def,domain_def,range_def,SUBSET_DEF,RES_FORALL,FORALL_PROD] THEN
            met []
          ) THEN
          pats `sc_fenced_sc_fences_heeded Q` [sc_fenced_sc_fences_heeded_def,RES_FORALL,RES_EXISTS] THEN          
          met []
        ),
        ( `(p_1,p_2) IN Xw.sc \/ (p_2,p_1) IN Xw.sc` by ALL_TAC THEN1
          ( pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,with_consume_relations_def,LET_DEF] THEN
            met []
          ) THEN
          `p_1 IN Xo.actions /\ p_2 IN Xo.actions /\ z IN Xo.actions /\ z' IN Xo.actions /\ z'' IN Xo.actions` by ALL_TAC THEN1
          ( pats `well_formed_rf Q` [well_formed_rf_def] THEN
            pats `consistent_mo Q` [consistent_mo_def] THEN
            pats `well_formed_threads Q` [well_formed_threads_def,RES_FORALL] THEN
            ps [relation_over_def,rel_over_def,domain_def,range_def,SUBSET_DEF,RES_FORALL,FORALL_PROD] THEN
            met []
          ) THEN
          pats `sc_fenced_sc_fences_heeded Q` [sc_fenced_sc_fences_heeded_def,RES_FORALL,RES_EXISTS] THEN          
          met []
        )
      ]
    )
  ]
)



val tc_alt_def = Q.store_thm ("tc_alt_def",
`!r. tch r = { xy | !tc'.
        (!x y.
           ((x,y) IN r) \/
           (?z. tc' (x,z) /\ tc' (z,y)) ==>
           tc' (x, y)) ==>
        tc' xy }`,
  fs [EXTENSION,tch_def,tc_def] THEN
  rw [] THEN
  FULL_SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) [IN_DEF]
)


val set_rel_to_rel_def = Define `set_rel_to_rel R = \x y. (x,y) IN R`

val rel_to_set_rel_def = Define `rel_to_set_rel R = {(x, y) | x, y | R x y}`


val set_rel_tc_eq_rel_tc = Q.store_thm ("set_rel_tc_eq_rel_t",
`!Rl. tch Rl = rel_to_set_rel (set_rel_to_rel Rl)^+`,
  fs [set_rel_to_rel_def,rel_to_set_rel_def,TC_DEF,tc_alt_def,EXTENSION] THEN
  FULL_SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) [] THEN
  met [CURRY_DEF,UNCURRY_DEF]
)



val tc_of_subset_then_subset = Q.store_thm ("tc_of_subset_then_subset",
`!A B. A SUBSET B ==> (tch A) SUBSET (tch B)`,
  GEN_TAC THEN
  GEN_TAC THEN
  fs [SUBSET_DEF,set_rel_tc_eq_rel_tc, rel_to_set_rel_def] THEN
  NTAC 2 STRIP_TAC THEN
  FULL_SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) [] THEN
  MATCH_MP_TAC TC_MONOTONE THEN
  fs [set_rel_to_rel_def]
)





val exists_vse_iff_w_hb = Q.store_thm ("exists_vse_iff_w_hb",
` !Xo Xw hb' rl r.
  let vse'   = visible_side_effect_set Xo.actions hb' in
  assumptions (Xo,Xw,("hb",hb')::rl) /\
  well_formed_threads (Xo,Xw,("hb",hb')::rl) /\
  finite_prefixes hb' Xo.actions /\
  relation_over Xo.actions hb' /\
  consistent_hb (Xo,Xw,("hb",hb')::rl) /\
  is_read r
  ==>
  ( (?w. w IN Xo.actions /\ (w,r) IN vse') <=>
    (?w. w IN Xo.actions /\ is_write w /\ (loc_of w = loc_of r) /\ (w,r) IN hb')
  )
`,
  rw [] THEN
  EQ_TAC THEN
  rw [] THEN
  Q.UNABBREV_TAC `vse'` THEN
  fs [visible_side_effect_set_def,RES_EXISTS] THENL
  [ met [],
    `r IN Xo.actions` by fs [relation_over_thm,RES_FORALL] THEN1 ts [] THEN
    `FINITE {a | a <> r /\ is_write a /\ (loc_of a = loc_of r) /\ (a,r) IN hb'}`
      by fs [finite_prefixes_def] THEN1
    ( `{a | a <> r /\ is_write a /\ (loc_of a = loc_of r) /\ (a,r) IN hb'} SUBSET
        {a | (a,r) IN hb'}` by fs [SUBSET_DEF,EXTENSION] THEN
        met [SUBSET_FINITE] 
    ) THEN
    `{a | a <> r /\ is_write a /\ (loc_of a = loc_of r) /\ (a,r) IN hb'} <> {}` by fs [EXTENSION] THEN1
    ( Q.EXISTS_TAC `w` THEN
        ts [RES_FORALL,consistent_hb_def,lem_is_irreflexive_def] THEN
        met [tc_rules]
    ) THEN
    `set_relation$acyclic hb'` by (ts [consistent_hb_def,set_relationTheory.acyclic_def,lem_is_irreflexive_def,RES_FORALL,herdTheory.irreflexive_def,tch_def] ) THEN
    `?x. x IN maximal_elements {a | a <> r /\ is_write a /\ (loc_of a = loc_of r) /\ (a,r) IN hb'} hb'` by met [finite_acyclic_has_maximal] THEN
    Q.EXISTS_TAC `x` THEN
    fs [maximal_elements_def] THEN
    `x IN Xo.actions` by fs [relation_over_thm,RES_FORALL] THEN1 ts [] THEN
    fs [] THEN
    CCONTR_TAC THEN
    fs [] THEN
    `x = c` by met [] THEN
    met [acyclic_def,tc_rules]
  ]
)





val lemHerdWellFormedThreads = store_thm("lemHerdWellFormedThreads",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  herd_well_formed_threads Xh.inits Xh.acts Xh.po Xh.data Xh.addr Xh.aswh
   Xh.lkh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD,well_formed_threads_def] THEN
  psg [herd_well_formed_threads_def] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  fs [well_formed_asw_def,well_formed_actions_def,RES_FORALL,lemHerdCondition_def,lemToHerdExecution_def,LET_DEF]
)



val exists_w_hb_then_vse = store_thm("exists_w_hb_then_vse",
``(!Xo Xw hb' rl r.
     (let vse' = visible_side_effect_set Xo.actions hb'
      in
        assumptions (Xo,Xw,("hb",hb')::rl) /\
        well_formed_threads (Xo,Xw,("hb",hb')::rl) /\
        finite_prefixes hb' Xo.actions /\
        relation_over Xo.actions hb' /\
        consistent_hb (Xo,Xw,("hb",hb')::rl) /\ is_read r /\
        ( ?w.
           w IN Xo.actions /\ is_write w /\ (loc_of w = loc_of r) /\
           (w,r) IN hb' )
        ==>
        (?w. w IN Xo.actions /\ (w,r) IN vse')
     )
)``,
  met [exists_vse_iff_w_hb]
)



val det_read_vse_then_rf = store_thm("det_read_vse_then_rf",
``( !Xo Xw hb' vse' ithb' sw' rs' hrs' dob' cad' r.
      det_read (Xo,Xw,[("hb",hb'); ("vse",vse'); ("ithb",ithb'); ("sw",sw');
                  ("rs",rs'); ("hrs",hrs'); ("dob",dob'); ("cad",cad')]) /\
      (r IN Xo.actions) /\ is_load r /\ (?w. (w IN Xo.actions) /\ (w,r) IN vse') ==>
        ?w'. w' IN Xo.actions /\ (w',r) IN Xw.rf
)``,
  ps ([det_read_def] @ qsimps) THEN
  met []
)


val acts_eq_actions = store_thm("acts_eq_actions",
``( (lemToHerdExecution Xo Xw addr data).acts = Xo.actions
)``,
  fs [lemToHerdExecution_def,LET_DEF]
)




val lemHerdRfForEveryRead = store_thm("lemHerdRfForEveryRead",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  rf_for_every_read Xh.acts Xh.rfh
)``,
  rw [] THEN
  `herd_well_formed_threads Xh.inits Xh.acts Xh.po Xh.data Xh.addr Xh.aswh
     Xh.lkh` by met [LET_DEF,lemHerdWellFormedThreads] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  psg [rf_for_every_read_def,RES_FORALL,RES_EXISTS] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  psg [lemToHerdExecution_def,LET_DEF] THEN
  fs [] THEN
  Q.ABBREV_TAC `Xh = lemToHerdExecution Xo Xw addr data` THEN
  pats `Q=Q''` [hblem,doblem,swlem,LET_DEF] THEN
  Q.ABBREV_TAC `rs' = release_sequence_set Xo.actions Xo.lk Xw.mo` THEN
  Q.ABBREV_TAC `hrs' = hypothetical_release_sequence_set Xo.actions Xo.lk
                  Xw.mo` THEN
  Q.ABBREV_TAC `sw' = release_acquire_fenced_synchronizes_with_set Xo.actions
               Xo.sb Xo.asw Xw.rf Xw.lo rs' hrs'` THEN
  Q.ABBREV_TAC `cad' = with_consume_cad_set Xo.actions Xo.sb Xo.dd Xw.rf` THEN
  Q.ABBREV_TAC `dob' = with_consume_dob_set Xo.actions Xw.rf rs' cad'` THEN
  Q.ABBREV_TAC `ithb' = inter_thread_happens_before Xo.actions Xo.sb sw' dob'` THEN
  Q.ABBREV_TAC `hb' = happens_before Xo.actions Xo.sb ithb'` THEN
  Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
  `relation_over Xo.actions hb'` by met [hb_over_actions,ithb_over_actions,LET_DEF] THEN
  ps [well_formed_threads_def,RES_FORALL,FORALL_PROD] THEN

  `?w. is_write w /\ (loc_of w = loc_of r) /\ (w,r) IN hb Xh` by ALL_TAC THEN1
  ( `?w. is_write w /\ (loc_of w = loc_of r) /\ (w,r) IN Xo.asw` by ALL_TAC THEN1
    ( `well_formed_addr Xh.acts Xh.addr Xh.po /\ well_formed_inits Xh.inits Xh.acts Xh.po` by ps [lemHerdCondition_def] THEN
      ps ([well_formed_addr_def,well_formed_inits_def] @ qsimps) THEN
      Q.UNABBREV_TAC `Xh` THEN
      fs [lemToHerdExecution_def,LET_DEF] THEN
      met []
    ) THEN
    `(w,r) IN hb Xh` by ALL_TAC THEN1
    ( `(w,r) IN sw'` by ALL_TAC THEN1
      ( Q.UNABBREV_TAC `sw'` THEN
        psg ([release_acquire_fenced_synchronizes_with_set_def,release_acquire_fenced_synchronizes_with_def] @ qsimps) THEN
        ps ([interthread_def] @ reloversimps @ qsimps) THEN
        met []
      ) THEN
      ps [happens_before_def,inter_thread_happens_before_def,LET_DEF] THEN
      met [IN_UNION,tc_rules]
    ) THEN
    met []
  ) THEN
  `?v. v IN Xo.actions /\ (v,r) IN vse'` by ALL_TAC THEN1
  ( `finite_prefixes (hb Xh) Xo.actions` by ALL_TAC THEN1
    ( ps [finite_prefixes_def,assumptions_def] THEN
      `!e. {e' | (e',e) IN hb Xh} SUBSET Xo.actions` by ALL_TAC THEN1
      ( fs ([SUBSET_DEF,LET_DEF] @ qsimps @ reloversimps) THEN
        met []
      ) THEN
      `FINITE Xh.acts` by met [lemHerdCondition_def] THEN
      Q.UNABBREV_TAC `Xh` THEN
      pats `Q` [LET_DEF,lemToHerdExecution_def] THEN
      fs [] THEN
      met [lemHerdCondition_def,SUBSET_FINITE]
    ) THEN
    `w IN Xo.actions` by ALL_TAC THEN1
    ( fs reloversimps THEN met [] ) THEN
    Q.UNABBREV_TAC `vse'` THEN
    MATCH_MP_TAC (SIMP_RULE std_ss ([LET_DEF] @ qsimps)exists_w_hb_then_vse) THEN
    Q.EXISTS_TAC `Xw` THEN
    Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
    Q.EXISTS_TAC `[("vse",vse'); ("ithb",ithb'); ("sw",sw');
                  ("rs",rs'); ("hrs",hrs'); ("dob",dob'); ("cad",cad')]` THEN
    fs [with_consume_relations_def,well_formed_threads_def,LET_DEF] THEN
    met []
  ) THEN
  Cases_on `is_load r` THEN1
  ( ps [with_consume_relations_def,LET_DEF] THEN
    met [acts_eq_actions,det_read_vse_then_rf]
  ) THEN
  `is_RMW r /\ is_write r` by ALL_TAC THEN1
  ( ps [is_write_def,is_load_def,is_read_def,is_RMW_def] THEN
    Cases_on `r` THEN
    fs []
  ) THEN
  `w <> r` by ALL_TAC THEN1
  ( pats `consistent_hb Q` ([consistent_hb_def,with_consume_relations_def,lem_is_irreflexive_def,LET_DEF] @ qsimps) THEN
    met [tc_rules]
  ) THEN
  `(w,r) IN Xw.mo` by ALL_TAC THEN1
  ( `is_at_atomic_location Xo.lk r` by ALL_TAC THEN1
    ( pats `actions_respect_location_kinds Xo.actions Xo.lk` ([actions_respect_location_kinds_def] @ qsimps) THEN
      specpat `!a.Q` `r` THEN
      fs ([is_at_atomic_location_def,is_RMW_def,loc_of_def] @ qsimps @ reloversimps) THEN
      Cases_on `r` THEN fs [] THEN met []
    ) THEN
    `(w,r) IN Xw.mo \/ (r,w) IN Xw.mo` by ALL_TAC THEN1
    ( pats `consistent_mo Q` ([consistent_mo_def] @ qsimps) THEN
      ps reloversimps THEN
      met []
    ) THEN
    pats `coherent_memory_use Q` ([coherent_memory_use_def,with_consume_relations_def,LET_DEF] @ qsimps) THEN
    met []
  ) THEN
  `?w'. adjacent_less_than Xw.mo Xo.actions w' r` by ALL_TAC THEN1
  ( `FINITE {a | (a,r) IN Xw.mo}` by ALL_TAC THEN1
    ( `FINITE Xh.acts` by met [lemHerdCondition_def] THEN
      Q.UNABBREV_TAC `Xh` THEN
      pats `Q` [LET_DEF,lemToHerdExecution_def] THEN
      fs [] THEN
      `{a | (a,r) IN Xw.mo} SUBSET Xo.actions` by ALL_TAC THEN1
      ( pats `consistent_mo Q` ([lem_is_irreflexive_def,consistent_mo_def] @ qsimps @ reloversimps) THEN
        fs [SUBSET_DEF] THEN
        met []
      ) THEN
      met [SUBSET_FINITE]
    ) THEN
    `set_relation$acyclic Xw.mo` by ALL_TAC THEN1
    ( pats `consistent_mo Q` ([lem_is_irreflexive_def,consistent_mo_def,with_consume_relations_def,LET_DEF] @ qsimps) THEN
      ps [set_relationTheory.acyclic_def,transitive_tc]
    ) THEN          
    `{a | (a,r) IN Xw.mo} <> {}` by ALL_TAC THEN1
    ( `w IN {a | (a,r) IN Xw.mo}` by fs [] THEN
      met [MEMBER_NOT_EMPTY]
    ) THEN
    `?x. x IN maximal_elements {a | (a,r) IN Xw.mo} Xw.mo` by met [finite_acyclic_has_maximal] THEN
    ps [maximal_elements_def] THEN
    Q.EXISTS_TAC `x` THEN
    ps ([adjacent_less_than_def] @ qsimps) THEN        
    pats `consistent_mo Q` ([lem_is_irreflexive_def,consistent_mo_def] @ qsimps) THEN
    met []
  ) THEN
  pats `rmw_atomicity Q` ([rmw_atomicity_def,with_consume_relations_def,LET_DEF] @ qsimps) THEN
  Q.EXISTS_TAC `w'` THEN
  `w' IN Xo.actions /\ r IN Xo.actions` by ALL_TAC THEN1
  ( pats `consistent_mo Q` ([consistent_mo_def,with_consume_relations_def,LET_DEF] @ qsimps @ reloversimps) THEN
    ps [adjacent_less_than_def] THEN
    met []
  ) THEN
  specpat `!a. Q` `r` THEN mppat `Q` THEN ps [] THEN st THEN
  specpat `!a. Q` `w'` THEN mppat `Q` THEN ps []
)




val lemHerdWellFormedRf = store_thm("lemHerdWellFormedRf",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  herd_well_formed_rf Xh.acts Xh.rfh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  fs ([LET_DEF,herd_well_formed_rf_def,well_formed_rf_def,lemToHerdExecution_def] @ qsimps) THEN
  met []
)




val lemHerdWellFormedCo = store_thm("lemHerdWellFormedCo",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  well_formed_co Xh.acts Xh.lkh Xh.co
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  fs ([LET_DEF,well_formed_co_def,consistent_mo_def,lemToHerdExecution_def,relation_over_def,lem_is_strict_partial_order_def] @ qsimps) THEN
  met []
)







val lemHerdWellFormedLo = store_thm("lemHerdWellFormedLo",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  well_formed_lo Xh.acts Xh.lkh Xh.loh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  fs ([LET_DEF,well_formed_lo_def,locks_only_consistent_lo_def,lemToHerdExecution_def,relation_over_def,lem_is_strict_partial_order_def,with_consume_relations_def] @ qsimps) THEN
  met []
)



val lemHerdHbAcyclic = store_thm("lemHerdHbAcyclic",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  hbAcyclic Xh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  pats `consistent_hb X`
    [LET_DEF,consistent_hb_def,with_consume_relations_def] THEN
  ps [LET_DEF,lemToHerdExecution_def,hbAcyclic_def,tch_def,herdTheory.irreflexive_def,herdTheory.acyclic_def] THEN
  pats `Q=Q''` [hblem,doblem,swlem,LET_DEF] THEN
  Q.ABBREV_TAC `rs' = release_sequence_set Xo.actions Xo.lk Xw.mo` THEN
  Q.ABBREV_TAC `hrs' = hypothetical_release_sequence_set Xo.actions Xo.lk
                  Xw.mo` THEN
  Q.ABBREV_TAC `sw' = release_acquire_fenced_synchronizes_with_set Xo.actions
               Xo.sb Xo.asw Xw.rf Xw.lo rs' hrs'` THEN
  Q.ABBREV_TAC `cad' = with_consume_cad_set Xo.actions Xo.sb Xo.dd Xw.rf` THEN
  Q.ABBREV_TAC `dob' = with_consume_dob_set Xo.actions Xw.rf rs' cad'` THEN
  Q.ABBREV_TAC `ithb' = inter_thread_happens_before Xo.actions Xo.sb sw' dob'` THEN
  Q.ABBREV_TAC `hb' = happens_before Xo.actions Xo.sb ithb'` THEN
  Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
  met []
)








val lemHerdCoherence = store_thm("lemHerdCoherence",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  coherence Xh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  psg ([coherence_def] @ herdsimps) THEN
  Q.ABBREV_TAC `hb'' = hb (lemToHerdExecution Xo Xw addr data)` THEN
  psg [lemToHerdExecution_def,LET_DEF] THEN fs [] THEN
  pats `coherent_memory_use Q` [coherent_memory_use_def,RES_EXISTS,EXISTS_PROD,LET_DEF,with_consume_relations_def] THEN
  pats `consistent_mo Q` ([consistent_mo_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  pats `well_formed_rf Q` ([well_formed_rf_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  pats `Q=Q''` [hblem,doblem,swlem,LET_DEF] THEN
  Q.ABBREV_TAC `rs' = release_sequence_set Xo.actions Xo.lk Xw.mo` THEN
  Q.ABBREV_TAC `hrs' = hypothetical_release_sequence_set Xo.actions Xo.lk
                  Xw.mo` THEN
  Q.ABBREV_TAC `sw' = release_acquire_fenced_synchronizes_with_set Xo.actions
               Xo.sb Xo.asw Xw.rf Xw.lo rs' hrs'` THEN
  Q.ABBREV_TAC `cad' = with_consume_cad_set Xo.actions Xo.sb Xo.dd Xw.rf` THEN
  Q.ABBREV_TAC `dob' = with_consume_dob_set Xo.actions Xw.rf rs' cad'` THEN
  Q.ABBREV_TAC `ithb' = inter_thread_happens_before Xo.actions Xo.sb sw' dob'` THEN
  Q.ABBREV_TAC `hb' = happens_before Xo.actions Xo.sb ithb'` THEN
  Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
  met []
)










val lemHerdRmwAtomicity = store_thm("lemHerdRmwAtomicity",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  rmwAtomicity Xh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  psg ([rmwAtomicity_def] @ herdsimps) THEN
  psg [lemToHerdExecution_def,LET_DEF] THEN fs [] THEN
  pats `consistent_mo Q` ([consistent_mo_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  pats `well_formed_rf Q` ([well_formed_rf_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  pats `rmw_atomicity Q` ([rmw_atomicity_def] @ qsimps) THEN
  CCONTR_TAC THEN
  ps [] THENL
  [ ( ps [adjacent_less_than_def] THEN
      met []
    ),
    ( `is_RMW z` by met [readWriteIsRmw] THEN
      `adjacent_less_than Xw.mo Xo.actions z' z` by met [] THEN
      ps ([adjacent_less_than_def] @ qsimps) THEN
      met []
    ),
    ( `is_RMW p_1` by met [readWriteIsRmw] THEN
      `adjacent_less_than Xw.mo Xo.actions z p_1` by met [] THEN
      ps ([adjacent_less_than_def,lem_is_irreflexive_def,transitive_def] @ qsimps) THEN
      met []
    )
  ]
)





val lemHerdAtomicRfConsistent = store_thm("lemHerdAtomicRfConsistent",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
  ==>
  atomicRfConsistent Xh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  psg ([atomicRfConsistent_def,hbarf_def,arf_def] @ herdsimps) THEN
  pats `Q=Q''` [hblem,doblem,swlem,LET_DEF] THEN
  Q.ABBREV_TAC `rs' = release_sequence_set Xo.actions Xo.lk Xw.mo` THEN
  Q.ABBREV_TAC `hrs' = hypothetical_release_sequence_set Xo.actions Xo.lk
                  Xw.mo` THEN
  Q.ABBREV_TAC `sw' = release_acquire_fenced_synchronizes_with_set Xo.actions
               Xo.sb Xo.asw Xw.rf Xw.lo rs' hrs'` THEN
  Q.ABBREV_TAC `cad' = with_consume_cad_set Xo.actions Xo.sb Xo.dd Xw.rf` THEN
  Q.ABBREV_TAC `dob' = with_consume_dob_set Xo.actions Xw.rf rs' cad'` THEN
  Q.ABBREV_TAC `ithb' = inter_thread_happens_before Xo.actions Xo.sb sw' dob'` THEN
  Q.ABBREV_TAC `hb' = happens_before Xo.actions Xo.sb ithb'` THEN
  Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
  Q.ABBREV_TAC `hb'' = hb (lemToHerdExecution Xo Xw addr data)` THEN
  psg [lemToHerdExecution_def,LET_DEF] THEN fs [] THEN
  rw [] THEN
  CCONTR_TAC THEN ps [] THEN
  pats `consistent_atomic_rf Q` [consistent_atomic_rf_def,RES_FORALL,FORALL_PROD,LET_DEF,with_consume_relations_def] THEN
  specpat `Q` `z` THEN specpat `Q` `p_1` THEN
  mppat `Q` THEN ps [] THEN CCONTR_TAC THEN
  `z IN Xo.actions /\ is_read p_1` by (fs
    [well_formed_rf_def,RES_FORALL,FORALL_PROD,LET_DEF,with_consume_relations_def]
    THEN
    met []
  ) THEN
  pats `rmw_atomicity Q` [rmw_atomicity_def,RES_FORALL,FORALL_PROD] THEN
  specpat `Q` `p_1` THEN mppat `Q` THEN ps [] THEN CCONTR_TAC THEN ps [] THEN
  specpat `Q` `z` THEN  mppat `Q` THEN ps [] THEN CCONTR_TAC THEN ps [] THEN1
  ( ps [is_load_def,is_read_def,is_RMW_def] THEN
    Cases_on `p_1` THEN fs []
  ) THEN
  ps [adjacent_less_than_def] THEN
  pats `coherent_memory_use Q` ([coherent_memory_use_def,RES_FORALL,FORALL_PROD,LET_DEF,with_consume_relations_def] @ qsimps) THEN
  met []
)









val lemHerdScConsistent = store_thm("lemHerdScConsistent",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
  ==>
  scConsistent Xh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  ps ([scConsistent_def,herdTheory.acyclic_def,tch_def,herdTheory.irreflexive_def,lem_is_irreflexive_def] @ qsimps) THEN
  rw [] THEN
  `(tch (scp (lemToHerdExecution Xo Xw addr data))) SUBSET (tch Xw.sc)`
  by met [tc_of_subset_then_subset] THEN
  ps [tch_def] THEN
  pats `sc_accesses_consistent_sc Q` [sc_accesses_consistent_sc_def,RES_FORALL,FORALL_PROD,RES_FORALL,FORALL_PROD,LET_DEF,with_consume_relations_def] THEN
  ps ([SUBSET_DEF,lem_is_irreflexive_def] @ qsimps) THEN
  `tc Xw.sc = Xw.sc` by met [transitive_tc] THEN
  met []
)









val lemHerdHbAgreesLo = store_thm("lemHerdHbAgreesLo",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
  ==>
  hbAgreesLo Xh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  ps [hbAgreesLo_def,herdTheory.irreflexive_def,seq_def,rcomp_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD] THEN
  pats `locks_only_consistent_lo Q` [locks_only_consistent_lo_def,with_consume_relations_def,LET_DEF,RES_FORALL] THEN
  pats `Q=Q''` [hblem,doblem,swlem,LET_DEF] THEN
  Q.ABBREV_TAC `rs' = release_sequence_set Xo.actions Xo.lk Xw.mo` THEN
  Q.ABBREV_TAC `hrs' = hypothetical_release_sequence_set Xo.actions Xo.lk
                  Xw.mo` THEN
  Q.ABBREV_TAC `sw' = release_acquire_fenced_synchronizes_with_set Xo.actions
               Xo.sb Xo.asw Xw.rf Xw.lo rs' hrs'` THEN
  Q.ABBREV_TAC `cad' = with_consume_cad_set Xo.actions Xo.sb Xo.dd Xw.rf` THEN
  Q.ABBREV_TAC `dob' = with_consume_dob_set Xo.actions Xw.rf rs' cad'` THEN
  Q.ABBREV_TAC `ithb' = inter_thread_happens_before Xo.actions Xo.sb sw' dob'` THEN
  Q.ABBREV_TAC `hb' = happens_before Xo.actions Xo.sb ithb'` THEN
  Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
  ps [lemToHerdExecution_def,LET_DEF] THEN fs [] THEN
  fs [rel_over_def,RES_FORALL,relation_over_def,range_def,domain_def,SUBSET_DEF] THEN
  met []
)





val lemHerdConsistentLo = store_thm("lemHerdConsistentLo",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
  ==>
  consistentLo Xh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  psg ([consistentLo_def,herdTheory.irreflexive_def,seq_def,rcomp_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD]
  @ herdsimps) THEN
  pats `locks_only_consistent_locks Q`
    [locks_only_consistent_locks_def,with_consume_relations_def,LET_DEF,RES_FORALL,FORALL_PROD,RES_EXISTS,EXISTS_PROD]
    THEN
  psg [lemToHerdExecution_def,LET_DEF] THEN fs [] THEN
  met []
)












val lemHerdNonAtomicRfConsistent = store_thm("lemHerdNonAtomicRfConsistent",
``(
!Xo Xw opsem herd_opsem p data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  lemHerdCondition herd_opsem opsem p /\
  opsem p Xo /\
  apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw) /\
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc)
==>
  nonAtomicRfConsistent Xh
)``,
  rw [] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  Q.UNABBREV_TAC `Xh` THEN
  rw [] THEN
  psg ([nonAtomicRfConsistent_def,is_empty_def,EXTENSION,LET_DEF] @ herdsimps) THEN
  Q.ABBREV_TAC `hb'' = hb (lemToHerdExecution Xo Xw addr data)` THEN
  psg [lemToHerdExecution_def,LET_DEF] THEN fs [] THEN
  rw [] THEN

  Cases_on `((p_1,p_2) NOTIN Xw.rf \/ p_2 NOTIN Xo.actions \/
 ~is_at_non_atomic_location Xo.lk p_2)` THEN
  fs [] THEN

  `is_write p_1 /\ p_1 IN Xo.actions /\ (loc_of p_1 = loc_of p_2) /\
  p_1 IN Xo.actions /\ is_read p_2` by ALL_TAC THEN1
  ( pats `well_formed_rf Q` [well_formed_rf_def,RES_FORALL,FORALL_PROD,LET_DEF,with_consume_relations_def] THEN
    met []
  ) THEN
  ps [] THEN
  pats `Q=Q''` [] THEN
  ps [hblem,doblem,swlem,LET_DEF] THEN
  Q.ABBREV_TAC `rs' = release_sequence_set Xo.actions Xo.lk Xw.mo` THEN
  Q.ABBREV_TAC `hrs' = hypothetical_release_sequence_set Xo.actions Xo.lk
                  Xw.mo` THEN
  Q.ABBREV_TAC `sw' = release_acquire_fenced_synchronizes_with_set Xo.actions
               Xo.sb Xo.asw Xw.rf Xw.lo rs' hrs'` THEN
  Q.ABBREV_TAC `cad' = with_consume_cad_set Xo.actions Xo.sb Xo.dd Xw.rf` THEN
  Q.ABBREV_TAC `dob' = with_consume_dob_set Xo.actions Xw.rf rs' cad'` THEN
  Q.ABBREV_TAC `ithb' = inter_thread_happens_before Xo.actions Xo.sb sw' dob'` THEN
  Q.ABBREV_TAC `hb' = happens_before Xo.actions Xo.sb ithb'` THEN
  Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
  `(p_1,p_2) IN vse'` by ALL_TAC
  THEN1
  ( pats `consistent_non_atomic_rf Q` [consistent_non_atomic_rf_def,RES_FORALL,FORALL_PROD,LET_DEF,with_consume_relations_def] THEN
    met []
  ) THEN
  `!p_1. (p_1,p_1) NOTIN tc hb'` by ALL_TAC THEN1
  ( pats `consistent_hb Q` [consistent_hb_def,RES_FORALL,FORALL_PROD,LET_DEF,with_consume_relations_def] THEN
    ps [lem_is_irreflexive_def,RES_FORALL,FORALL_PROD] THEN
    met []
  ) THEN
  Q.UNABBREV_TAC `vse'` THEN
  pats `(p_1,p_2) IN Q` [visible_side_effect_set_def,RES_EXISTS,EXISTS_PROD] THEN
  ps [] THEN
  rw [] THEN
  specpat `!c. Q` `z` THEN
  ps [] THEN
  met [tc_rules]
)







val lemHerdExistsConsistentExecution = store_thm("lemHerdExistsConsistentExecution",
``(
!Xo Xw opsem herd_opsem p.
lemHerdCondition herd_opsem opsem p /\
opsem p Xo /\
apply_tree with_consume_consistent_execution
        (Xo,Xw,with_consume_relations Xo Xw)
==>
?data addr.
  let Xh = lemToHerdExecution Xo Xw addr data in
  herd_opsem p Xh /\
  (tc (data UNION addr) = Xo.dd) /\
  (hblem Xo Xw = hb Xh) /\
  (scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc) /\
  cpp11$provides_clauses Xh
)``,
  rw [] THEN
  ` ?data addr.
    let Xh = lemToHerdExecution Xo Xw addr data in
    herd_opsem p Xh /\
    (tc (data UNION addr) = Xo.dd) /\
    (hblem Xo Xw = hb Xh) /\
    scp (lemToHerdExecution Xo Xw addr data) SUBSET Xw.sc` by met [lemHerdExistsEquivalentExecution] THEN
  Q.EXISTS_TAC `data` THEN
  Q.EXISTS_TAC `addr` THEN
  ps [LET_DEF] THEN
  rw [] THEN1 met [] THEN1 met [] THEN
  ps [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def] THEN
  `herd_well_formed_threads Xh.inits Xh.acts Xh.po Xh.data Xh.addr Xh.aswh
   Xh.lkh` by met [LET_DEF,lemHerdWellFormedThreads] THEN
  `rf_for_every_read Xh.acts Xh.rfh` by met [LET_DEF,lemHerdRfForEveryRead] THEN
  `herd_well_formed_rf Xh.acts Xh.rfh` by met [LET_DEF,lemHerdWellFormedRf] THEN
  `well_formed_co Xh.acts Xh.lkh Xh.co` by met [LET_DEF,lemHerdWellFormedCo] THEN
  `well_formed_lo Xh.acts Xh.lkh Xh.loh` by met [LET_DEF,lemHerdWellFormedLo] THEN
  `hbAcyclic Xh` by met [LET_DEF,lemHerdHbAcyclic] THEN
  `rmwAtomicity Xh` by met [LET_DEF,lemHerdRmwAtomicity] THEN
  `atomicRfConsistent Xh` by met [LET_DEF,lemHerdAtomicRfConsistent] THEN
  `scConsistent Xh` by met [LET_DEF,lemHerdScConsistent] THEN
  `hbAgreesLo Xh` by met [LET_DEF,lemHerdHbAgreesLo] THEN
  `consistentLo Xh` by met [LET_DEF,lemHerdConsistentLo] THEN
  `nonAtomicRfConsistent Xh` by met [LET_DEF,lemHerdNonAtomicRfConsistent] THEN
  `coherence Xh` by met [LET_DEF,lemHerdCoherence] THEN
  ps []
)










val strictPartialExistsPartial = store_thm("strictPartialExistsPartial",
``(
!r s.
  let r' = r UNION {(a,a) | a IN s} in
  rel_over r s /\ lem_is_strict_partial_order r
   ==>
   partial_order r' s /\  r SUBSET r'
)``,
  rw [] THEN
  `r SUBSET r'` by met [IN_UNION,SUBSET_DEF] THEN
  ps [partial_order_def,lem_is_strict_partial_order_def,irreflexive_def,RES_FORALL,rel_over_def] THEN
  rw [] THENL
  [ ( Q.UNABBREV_TAC `r'` THEN
      ps [SUBSET_DEF,domain_def,FORALL_PROD,UNION_DEF,CROSS_DEF,FST,SND] THEN
      met []
    ),
    ( Q.UNABBREV_TAC `r'` THEN
      ps [SUBSET_DEF,range_def,FORALL_PROD,UNION_DEF,CROSS_DEF,FST,SND] THEN
      met []
    ),
    ( Q.UNABBREV_TAC `r'` THEN
      ps [transitive_def,domain_def,range_def,SUBSET_DEF] THEN
      met []
    ),
    ( Q.UNABBREV_TAC `r'` THEN
      ps [reflexive_def]
    ),
    ( Q.UNABBREV_TAC `r'` THEN
      ps [antisym_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD,transitive_def] THEN
      met []
    )
  ]
)



val partialToTotal = store_thm("partialToTotal",
``(
!r s.
  countable s /\
  rel_over r s /\
  lem_is_strict_partial_order r /\
  finite_prefixes r s ==>
   ?r'. rel_over r' s /\ lem_is_strict_total_order_on r' s /\  r SUBSET r'
)``,
  rw [] THEN
  Q.ABBREV_TAC `r' = r UNION {(a,a) | a IN s}` THEN
  `partial_order r' s /\  r SUBSET r'` by met[strictPartialExistsPartial,LET_DEF] THEN
  `rel_over r' s` by ALL_TAC THEN1
  ( Q.UNABBREV_TAC `r'` THEN
    ps [rel_over_def,domain_def,range_def,SUBSET_DEF,UNION_DEF] THEN
    met []
  ) THEN
  `finite_prefixes r' s` by ALL_TAC THEN1
  ( ps [finite_prefixes_def] THEN
    rw [] THEN
    `{e' | (e',e) IN r'} = (e INSERT {e' | (e',e) IN r})` by ALL_TAC THEN1
    ( Q.UNABBREV_TAC `r'` THEN
      ps [INSERT_DEF,EXTENSION] THEN
      met []
    ) THEN
    met [FINITE_INSERT]
  ) THEN
  `?r''. linear_order r'' s /\ r' SUBSET r''` by met [linear_order_of_countable_po] THEN
  Q.ABBREV_TAC `r''' = strict r''` THEN
  Q.EXISTS_TAC `r'''` THEN
  `rel_over r''' s` by ALL_TAC THEN1
  ( Q.UNABBREV_TAC `r'''` THEN
    ps [rel_over_def,linear_order_def,strict_def,domain_def,range_def,SUBSET_DEF] THEN
    met []
  ) THEN
  `r SUBSET r'''` by ALL_TAC THEN1
  ( Q.UNABBREV_TAC `r'''` THEN
    psg [SUBSET_DEF,strict_def,FORALL_PROD] THEN
    rw [] THEN1 met [SUBSET_DEF] THEN
    ps [lem_is_strict_partial_order_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD] THEN
    met []
  ) THEN
  Q.UNABBREV_TAC `r'''` THEN
  ps [lem_is_strict_total_order_on_def,linear_order_def,lem_is_strict_partial_order_on_def] THEN
  ps [lem_is_irreflexive_def,set_relationTheory.irreflexive_def,strict_def,RES_FORALL,FORALL_PROD,transitive_def,lem_is_trichotomous_on_def,antisym_def,lem_transitive_on_def] THEN
  met []
)




val herdLemExistsTotalSc = store_thm("herdLemExistsTotalSc",
``(
!Xh opsem herd_opsem p.
lemHerdCondition herd_opsem opsem p /\
herd_opsem p Xh /\
cpp11$provides_clauses Xh
==>
?sct. tc (scp Xh) SUBSET sct /\ 
      rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
      lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a}
)``,
  rw [] THEN
  `rel_over (scp Xh) {a | a IN Xh.acts /\ is_seq_cst a}` by ALL_TAC THEN1
  ( ps mysimps THEN
    ps [rel_over_def,SUBSET_DEF,range_def,domain_def] THEN
    met []
  ) THEN
  `rel_over (tc (scp Xh)) {a | a IN Xh.acts /\ is_seq_cst a}` by ALL_TAC THEN1
  ( ps [rel_over_def,SUBSET_DEF,range_def,domain_def] THEN
    met [tc_cases_right,tc_cases_left]
  ) THEN
  `lem_is_strict_partial_order (tc (scp Xh))` by ALL_TAC THEN1
  ( ps [lem_is_strict_partial_order_def,lem_is_irreflexive_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,scConsistent_def,herdTheory.acyclic_def,tch_def,herdTheory.irreflexive_def] THEN
    met [tc_transitive]
  ) THEN
  `countable {a | a IN Xh.acts /\ is_seq_cst a} /\ finite_prefixes (tc (scp Xh)) {a | a IN Xh.acts /\ is_seq_cst a}` by ALL_TAC THEN1
  ( `FINITE Xh.acts` by ps [lemHerdCondition_def] THEN
    `{a | a IN Xh.acts /\ is_seq_cst a} SUBSET Xh.acts` by fs [SUBSET_DEF] THEN
    `FINITE {a | a IN Xh.acts /\ is_seq_cst a}` by met [SUBSET_FINITE] THEN
    rw [] THEN1 met [finite_countable] THEN
    ps [finite_prefixes_def] THEN
    rw [] THEN
    `{e' | (e',e) IN tc (scp Xh)} SUBSET {a | a IN Xh.acts /\ is_seq_cst a}` by ALL_TAC THEN1
    ( ps [rel_over_def,SUBSET_DEF,range_def,domain_def] THEN
      met []
    ) THEN
    met [SUBSET_FINITE]
  ) THEN
  met [partialToTotal]
)





val herd_consistent_then_hb_requirements = store_thm("herd_consistent_then_hb_requirements",
``(
!Xh.
  cpp11$provides_clauses Xh ==>
  (well_formed_actions Xh.acts) /\
  (!a b. (a,b) IN pp_asw Xh ==> (tid_of a <> tid_of b)) /\
  (!a b. a IN Xh.inits /\ b IN Xh.acts /\ b NOTIN Xh.inits ==> (tid_of a <> tid_of b)) /\
  Xh.rfh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.loh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.po  SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (cross Xh.acts Xh.acts) /\
  dd Xh SUBSET (cross Xh.acts Xh.acts) /\
  Xh.co  SUBSET (WW Xh) /\
  Xh.rfh SUBSET (WR Xh) /\
  Xh.aswh  SUBSET (cross Xh.acts Xh.acts)
)``,
  STRIP_TAC THEN
  ps [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def] THEN
  rw [] THENL
  [ ps [herd_provides_def,herd_well_formed_threads_def],
    ( ps [herd_provides_def,herd_well_formed_threads_def,well_formed_asw_def,pp_asw_def,well_formed_inits_def,RES_FORALL,EXTENSION,FORALL_PROD] THEN
      met []
    ),
    ( ps [herd_provides_def,herd_well_formed_threads_def,well_formed_inits_def,RES_FORALL] THEN
      met []
    ),
    ( ps [herd_provides_def,herd_well_formed_threads_def,herd_well_formed_rf_def,RES_FORALL,RES_FORALL,SUBSET_DEF] THEN
      rw [] THEN
      specpat `!a. Q` `x` THEN
      Cases_on `x` THEN mppat `Q` THEN ps [] THEN rw [] THEN
      ps [cross_def]
    ),
    ( ps [herd_provides_def,herd_well_formed_threads_def,well_formed_lo_def,RES_FORALL,RES_FORALL,SUBSET_DEF] THEN
      rw [] THEN
      Cases_on `x` THEN
      ps [rel_over_def,cross_def,SUBSET_DEF,domain_def,range_def] THEN
      met []
    ),
    ( ps [herd_provides_def,herd_well_formed_threads_def,well_formed_po_def,RES_FORALL,RES_FORALL,SUBSET_DEF] THEN
      rw [] THEN
      Cases_on `x` THEN
      ps [rel_over_def,cross_def,SUBSET_DEF,domain_def,range_def] THEN
      met []
    ),
    ( ps [herd_provides_def,herd_well_formed_threads_def,well_formed_co_def,RES_FORALL,RES_FORALL,SUBSET_DEF] THEN
      rw [] THEN
      Cases_on `x` THEN
      ps [rel_over_def,cross_def,SUBSET_DEF,domain_def,range_def] THEN
      met []
    ),
    ( ps [herd_provides_def,herd_well_formed_threads_def,well_formed_po_def,well_formed_addr_def,well_formed_data_def,RES_FORALL,RES_FORALL,SUBSET_DEF,dd_def,tch_def] THEN
      rw [] THEN
      Cases_on `x` THEN
      ps [rel_over_def,cross_def,SUBSET_DEF,domain_def,range_def] THEN
      ps qsimps THEN
      `?j. (q,j) IN (Xh.data UNION Xh.addr)` by met [tc_cases_left] THEN
      `?j. (j,r) IN (Xh.data UNION Xh.addr)` by met [tc_cases_right] THEN
      met [IN_UNION]
    ),
    ( ps [herd_provides_def,herd_well_formed_threads_def,well_formed_co_def,RES_FORALL,RES_FORALL,SUBSET_DEF,WW_def,cross_def,W_def] THEN
      rw [] THEN
      Cases_on `x` THEN
      rw [] THEN
      ps [rel_over_def,cross_def,SUBSET_DEF,domain_def,range_def] THEN
      met []
    ),
    ( ps [herd_provides_def,herd_well_formed_threads_def,herd_well_formed_rf_def,RES_FORALL,RES_FORALL,SUBSET_DEF,WR_def,cross_def,W_def,R_def] THEN
      rw [] THEN
      specpat `!a. Q` `x` THEN
      Cases_on `x` THEN mppat `Q` THEN ps [] THEN rw []
    ),
    ( ps [herd_provides_def,
          herd_well_formed_threads_def,
          well_formed_asw_def,
          well_formed_inits_def,
          RES_FORALL,
          FORALL_PROD,
          cross_def,
          SUBSET_DEF,
          EXTENSION]
    )
  ]
)




val herd_consistent_then_partial_eq = store_thm("herd_consistent_then_partial_eq",
``(
!Xh sct p herd_opsem opsem.
  let Xo = herdToLemPreExecution Xh in
  let Xw = herdToLemExecutionWitness Xh sct in
  cpp11$provides_clauses Xh
  ==>
  lemHerdPartialExecutionEquivalent (Xo,Xw) Xh
)``,
  rw [] THEN
  Q.UNABBREV_TAC `Xo` THEN
  Q.UNABBREV_TAC `Xw` THEN
  psg [herdToLemPreExecution_def,herdToLemExecutionWitness_def,pp_asw_def,dd_def,lemHerdPartialExecutionEquivalent_def,lemHerdPreExecutionEquivalent_def,lemToHerdExecution_def,compr_def,rcomp_def,seq_def,EXTENSION] THEN
  fs [LET_DEF,FORALL_PROD,UNION_DEF,EXTENSION] THEN
  `well_formed_inits Xh.inits Xh.acts Xh.po` by met [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,herd_well_formed_threads_def] THEN
  `well_formed_asw Xh.acts Xh.inits Xh.aswh` by met [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,herd_well_formed_threads_def] THEN
  ps [well_formed_asw_def,well_formed_inits_def,RES_FORALL,RES_EXISTS] THEN
  fs [lemToHerdExecution_def,LET_DEF,SUBSET_DEF] THEN
  met []
)



val herd_consistent_then_hb_eq = store_thm("herd_consistent_then_hb_eq",
``(
!Xh sct p herd_opsem opsem.
  let Xo = herdToLemPreExecution Xh in
  let Xw = herdToLemExecutionWitness Xh sct in
  cpp11$provides_clauses Xh
  ==>
  (hblem Xo Xw = hb Xh)
)``,
  rw [] THEN
  `lemHerdPartialExecutionEquivalent (Xo,Xw) Xh` by met [herd_consistent_then_partial_eq,LET_DEF] THEN
  ASSUME_TAC herd_consistent_then_hb_requirements THEN
  fs [LET_DEF] THEN
  RES_TAC THEN
  MATCH_MP_TAC hbEquivalent THEN
  ps [] THEN
  met [hbEquivalent]
)






val herdLemExistsEquivalentExecution = store_thm("herdLemExistsEquivalentExecution",
``(
!Xh opsem herd_opsem p.
lemHerdCondition herd_opsem opsem p /\
herd_opsem p Xh /\
cpp11$provides_clauses Xh
==>
?sct.
  tc (scp Xh) SUBSET sct /\ 
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  let Xo = herdToLemPreExecution Xh in
  let Xw = herdToLemExecutionWitness Xh sct in
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
)``,
  rw [] THEN
  `?sct. tc (scp Xh) SUBSET sct /\ 
      rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
      lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a}` by met [herdLemExistsTotalSc] THEN
  Q.EXISTS_TAC `sct` THEN
  ps [LET_DEF] THEN
  rw [] THEN1 met [lemHerdCondition_def] THEN
  ASSUME_TAC herd_consistent_then_hb_eq THEN
  fs [LET_DEF] THEN
  met []
)





val lemHerdAssumptions = store_thm("lemHerdAssumptions",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  assumptions (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  ps [finite_prefixes_def,assumptions_def] THEN
  `(!e. {e' | (e',e) IN Xw.rf} SUBSET Xh.acts) /\
   (!e. {e' | (e',e) IN Xw.sc} SUBSET Xh.acts) /\
   (!e. {e' | (e',e) IN Xw.mo} SUBSET Xh.acts) /\
   (!e. {e' | (e',e) IN Xw.lo} SUBSET Xh.acts)
  ` by ALL_TAC THEN1
  ( Q.UNABBREV_TAC `Xw` THEN
    fs ([SUBSET_DEF,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,well_formed_co_def,well_formed_lo_def,herd_well_formed_rf_def,herdToLemExecutionWitness_def,LET_DEF] @ qsimps @ reloversimps) THEN
    met []
  ) THEN
  met [lemHerdCondition_def,SUBSET_FINITE]
)






val lemHerdWFT = store_thm("lemHerdWFT",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  well_formed_threads (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  `herd_well_formed_threads Xh.inits Xh.acts Xh.po Xh.data Xh.addr Xh.aswh Xh.lkh` by ps [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def] THEN
  ps [well_formed_threads_def,herd_well_formed_threads_def] THEN
  Q.UNABBREV_TAC `Xo` THEN
  fs [herdToLemPreExecution_def,LET_DEF] THEN
  ps [well_formed_actions_def] THEN
  rw [] THENL
  [ ps [well_formed_po_def,relation_over_def],
    ( ps [EXTENSION,SUBSET_DEF,well_formed_asw_def,well_formed_inits_def] THEN
      psg (reloversimps @ herdsimps) THEN   
      met []
    ),
    ( ps ([well_formed_po_def,threadwise_def] @ qsimps @ reloversimps) THEN
      met []
    ),
    ( psg herdsimps THEN
      ps ([interthread_def,well_formed_asw_def,well_formed_inits_def] @ qsimps @ reloversimps) THEN
      met []
    ),
    ps [well_formed_po_def],
    ( psg herdsimps THEN
      ps ([lem_is_irreflexive_def,well_formed_data_def,well_formed_addr_def,well_formed_po_def,lem_is_strict_partial_order_def,SUBSET_DEF] @ qsimps) THEN
      `tc {x | x IN Xh.data \/ x IN Xh.addr} SUBSET Xh.po` by ALL_TAC THEN1
      ( `{x | x IN Xh.data \/ x IN Xh.addr} SUBSET Xh.po` by ALL_TAC THEN1
        ( fs (qsimps @ [SUBSET_DEF]) THEN
          met []
        ) THEN
        met [transitive_tc,tc_of_subset_then_subset,tch_def]
      ) THEN
      met [tc_transitive,SUBSET_DEF,transitive_def]
    ),
    ( psg herdsimps THEN
      ps ([well_formed_data_def,well_formed_addr_def,well_formed_po_def,SUBSET_DEF,lem_is_strict_partial_order_def] @ qsimps) THEN
      `tc {x | x IN Xh.data \/ x IN Xh.addr} SUBSET Xh.po` by ALL_TAC THEN1
      ( `{x | x IN Xh.data \/ x IN Xh.addr} SUBSET Xh.po` by ALL_TAC THEN1
        ( fs (qsimps @ [SUBSET_DEF]) THEN
          met []
        ) THEN
        met [transitive_tc,tc_of_subset_then_subset,tch_def]
      ) THEN
      met [SUBSET_DEF]
    ),
    fs
    ([indeterminate_sequencing_def,lem_is_irreflexive_def,well_formed_po_def]
    @ qsimps),
    ( fs ([lem_is_irreflexive_def,well_formed_po_def,well_formed_asw_def,well_formed_inits_def,sbasw_def] @ qsimps) THEN
      CCONTR_TAC THEN ps [] THEN          
      `tc (Xh.po UNION pp_asw Xh) SUBSET (Xh.po UNION pp_asw Xh)` by ALL_TAC THEN1
      ( `!x y. (?z. (x,z) IN Xh.po UNION pp_asw Xh /\ (z,y) IN Xh.po UNION pp_asw Xh) ==> (x,y) IN Xh.po UNION pp_asw Xh` by ALL_TAC THEN1
        ( psg herdsimps THEN
          rw [] THEN
          ps (reloversimps @ [EXTENSION] @ qsimps) THEN
          met [lem_is_strict_partial_order_def,transitive_def]
        ) THEN
        ASSUME_TAC (Q.ISPEC `\x (y:action). (x,y) IN (Xh.po UNION pp_asw Xh)` (Q.ISPEC `Xh.po UNION pp_asw Xh` tc_ind_left)) THEN
        fs ([SUBSET_DEF] @ qsimps) THEN
        met []
      ) THEN
      `(p_1,p_1) IN (Xh.po UNION pp_asw Xh)` by met [SUBSET_DEF] THEN
      pats `Q` [UNION_DEF] THEN
      pats `Q` herdsimps THEN
      pats `Xh.aswh = Q` ([EXTENSION] @ qsimps) THEN
      ps ([lem_is_strict_partial_order_def,lem_is_irreflexive_def] @ qsimps) THEN
      met []
    ),
    ( fs [finite_prefixes_def,assumptions_def,sbasw_def] THEN
      `!e. {e' | (e',e) IN tc (Xh.po UNION pp_asw Xh)} SUBSET Xh.acts
      ` by ALL_TAC THEN1
      ( fs ([SUBSET_DEF,well_formed_po_def,well_formed_asw_def,well_formed_inits_def,LET_DEF] @ qsimps @ reloversimps) THEN
        rw [] THEN
        `?e'. (x,e') IN (Xh.po UNION pp_asw Xh)` by met [tc_cases_left] THEN
        pats `Q` herdsimps THEN
        ps [EXTENSION] THEN
        met []
      ) THEN
      met [lemHerdCondition_def,SUBSET_FINITE]
    )
  ]
)



val lemHerdWFrf = store_thm("lemHerdWFrf",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  well_formed_rf (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  Q.UNABBREV_TAC `Xw` THEN
      Q.UNABBREV_TAC `Xo` THEN
      fs ([SUBSET_DEF,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,herd_well_formed_rf_def,well_formed_rf_def,herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
      met []
)


val lemHerdConsistentLocks = store_thm("lemHerdConsistentLocks",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  locks_only_consistent_locks (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  Q.UNABBREV_TAC `Xw` THEN
      Q.UNABBREV_TAC `Xo` THEN
      fs ([SUBSET_DEF,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,consistentLo_def,locks_only_consistent_locks_def,herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF,well_formed_lo_def] @ qsimps @ reloversimps) THEN
      pats `irreflexive (mv Xh)` herdsimps THEN
      met []
)




val lemHerdConsistentlo = store_thm("lemHerdConsistentlo",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  locks_only_consistent_lo (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([SUBSET_DEF,lem_is_strict_partial_order_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,hbAgreesLo_def,locks_only_consistent_lo_def,LET_DEF,well_formed_lo_def,with_consume_relations_def] @ qsimps @ reloversimps) THEN
      Q.UNABBREV_TAC `Xw` THEN
      Q.UNABBREV_TAC `Xo` THEN
      fs ([herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
      pats `irreflexive (seq Xh.loh (hb Xh))` herdsimps THEN
      fs [hblem,doblem,swlem,LET_DEF] THEN
      met []
)



val lemHerdConsistentmo = store_thm("lemHerdConsistentmo",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  consistent_mo (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([SUBSET_DEF,lem_is_strict_partial_order_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,consistent_mo_def,LET_DEF,well_formed_co_def,with_consume_relations_def] @ qsimps @ reloversimps) THEN
      Q.UNABBREV_TAC `Xw` THEN
      Q.UNABBREV_TAC `Xo` THEN
      fs ([herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps)
)




val lemHerdConsistentsc = store_thm("lemHerdConsistentsc",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  sc_accesses_consistent_sc (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([SUBSET_DEF,lem_is_strict_partial_order_on_def,lem_is_strict_total_order_on_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,sc_accesses_consistent_sc_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  Q.UNABBREV_TAC `Xw` THEN
  Q.UNABBREV_TAC `Xo` THEN
  fs ([herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
  fs [hblem,doblem,swlem,LET_DEF] THEN
  rw [] THENL
  [ ( fs ([lem_transitive_on_def,transitive_def] @ qsimps) THEN
      met []
    ),
    ( fs ([set_relationTheory.irreflexive_def,lem_is_irreflexive_def] @ qsimps) THEN
      met []
    ),
    ( CCONTR_TAC THEN ps [] THEN
      `(b,a) IN scp Xh` by ALL_TAC THEN1
      ( psg herdsimps THEN
        met []
      ) THEN
      `(b,a) IN sct` by met [tc_rules] THEN
      ps ([lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
      met []
    ),
    ( CCONTR_TAC THEN ps [] THEN
      `(b,a) IN scp Xh` by ALL_TAC THEN1
      ( psg herdsimps THEN
        met []
      ) THEN
      `(b,a) IN sct` by met [tc_rules] THEN
      ps ([lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
      met []
    ),
    ( ps ([lem_is_trichotomous_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
      met []
    )
  ]
)





val lemHerdConsistentscFH = store_thm("lemHerdConsistentscFH",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  sc_fenced_sc_fences_heeded (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([SUBSET_DEF,lem_is_strict_partial_order_on_def,lem_is_strict_total_order_on_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,sc_fenced_sc_fences_heeded_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
      Q.UNABBREV_TAC `Xw` THEN
      Q.UNABBREV_TAC `Xo` THEN
      fs ([herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
      fs [hblem,doblem,swlem,LET_DEF] THEN
      CCONTR_TAC THEN ps [] THENL
      [ ( `(f,w') IN scp Xh` by ALL_TAC THEN1
          ( psg herdsimps THEN
            `f <> w'` by ALL_TAC THEN1
            ( fs ([set_relationTheory.irreflexive_def] @ qsimps) THEN
              met []
            ) THEN
            met []
          ) THEN
          ps ([lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
          met [tc_rules]
        ),
        ( `(r,f) IN scp Xh` by ALL_TAC THEN1
          ( psg herdsimps THEN
            `r <> f` by ALL_TAC THEN1
            ( fs ([set_relationTheory.irreflexive_def] @ qsimps) THEN
              met []
            ) THEN
            met []
          ) THEN
          ps ([lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
          met [tc_rules]
        ),
        ( `(f',f) IN scp Xh` by ALL_TAC THEN1
          ( psg herdsimps THEN
            `f' <> f` by ALL_TAC THEN1
            ( fs ([set_relationTheory.irreflexive_def] @ qsimps) THEN
              met []
            ) THEN
            met []
          ) THEN
          ps ([lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
          met [tc_rules]
        ),
        ( `(f',f) IN scp Xh` by ALL_TAC THEN1
          ( psg herdsimps THEN
            `f' <> f` by ALL_TAC THEN1
            ( fs ([set_relationTheory.irreflexive_def] @ qsimps) THEN
              met []
            ) THEN
            met []
          ) THEN
          ps ([lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
          met [tc_rules]
        ),
        ( `(f,w') IN scp Xh` by ALL_TAC THEN1
          ( psg herdsimps THEN
            `f <> w'` by ALL_TAC THEN1
            ( fs ([set_relationTheory.irreflexive_def] @ qsimps) THEN
              met []
            ) THEN
            met []
          ) THEN
          ps ([lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
          met [tc_rules]
        ),
        ( `(w,f) IN scp Xh` by ALL_TAC THEN1
          ( psg herdsimps THEN
            `w <> f` by ALL_TAC THEN1
            ( fs ([set_relationTheory.irreflexive_def] @ qsimps) THEN
              met []
            ) THEN
            met []
          ) THEN
          ps ([lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
          met [tc_rules]
        )
      ]
)




val lemHerdConsistenthb = store_thm("lemHerdConsistenthb",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  consistent_hb (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([SUBSET_DEF,lem_is_strict_partial_order_on_def,lem_is_strict_total_order_on_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,hbAcyclic_def,consistent_hb_def,LET_DEF,with_consume_relations_def,herdTheory.acyclic_def,tch_def,herdTheory.irreflexive_def] @ qsimps @ reloversimps) THEN
      Q.UNABBREV_TAC `Xw` THEN
      Q.UNABBREV_TAC `Xo` THEN
      fs ([herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
      fs [hblem,doblem,swlem,LET_DEF]
)





val lemHerdDetRead= store_thm("lemHerdDetRead",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  det_read (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([herd_provides_def,rf_for_every_read_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,det_read_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  Q.UNABBREV_TAC `Xw` THEN
  Q.UNABBREV_TAC `Xo` THEN
  fs ([hblem,doblem,swlem,LET_DEF,herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
  rw [] THEN
  `is_read r` by ALL_TAC THEN1
  ( fs [is_read_def,is_load_def] THEN
    Cases_on `r` THEN fs []
  ) THEN
  EQ_TAC THEN1 met [] THEN
  rw [] THEN
  `?w. is_write w /\ (loc_of w = loc_of r) /\ (w,r) IN hb Xh` by ALL_TAC THEN1
  ( ps [herd_well_formed_threads_def] THEN
    `?w. is_write w /\ (loc_of w = loc_of r) /\ (w,r) IN Xo.asw` by ALL_TAC THEN1
    ( `well_formed_addr Xh.acts Xh.addr Xh.po /\ well_formed_inits Xh.inits Xh.acts Xh.po` by ps [lemHerdCondition_def] THEN
      ps ([well_formed_addr_def,well_formed_inits_def] @ qsimps) THEN
      met []
    ) THEN
    `(w,r) IN hb Xh` by ALL_TAC THEN1
    ( `(w,r) IN (release_acquire_fenced_synchronizes_with_set Xh.acts Xh.po
                   (pp_asw Xh) Xh.rfh Xh.loh
                   (release_sequence_set Xh.acts Xh.lkh Xh.co)
                   (hypothetical_release_sequence_set Xh.acts Xh.lkh Xh.co))` by ALL_TAC THEN1
      ( psg ([release_acquire_fenced_synchronizes_with_set_def,release_acquire_fenced_synchronizes_with_def] @ qsimps) THEN
        ps ([well_formed_asw_def,well_formed_inits_def] @ qsimps @ reloversimps) THEN
        met []
      ) THEN
      `(w,r) IN (inter_thread_happens_before Xh.acts Xh.po
                   (release_acquire_fenced_synchronizes_with_set Xh.acts Xh.po
               (pp_asw Xh) Xh.rfh Xh.loh
               (release_sequence_set Xh.acts Xh.lkh Xh.co)
               (hypothetical_release_sequence_set Xh.acts Xh.lkh Xh.co))
            (with_consume_dob_set Xh.acts Xh.rfh
               (release_sequence_set Xh.acts Xh.lkh Xh.co)
               (with_consume_cad_set Xh.acts Xh.po (dd Xh) Xh.rfh)))` by met [ IN_UNION,tc_rules,inter_thread_happens_before_def,LET_DEF] THEN
      ps [happens_before_def] THEN
      met [IN_UNION]
    ) THEN
    met []
  ) THEN
  `!a b. (a,b) IN hb Xh ==> a IN Xh.acts /\ b IN Xh.acts` by ALL_TAC THEN1
  ( psg ([hb_def,ithb_def,ithbr_def] @ herdsimps) THEN
    ps ([herd_well_formed_threads_def,well_formed_asw_def,well_formed_inits_def,well_formed_po_def,SUBSET_DEF,well_formed_co_def,well_formed_lo_def] @ reloversimps) THEN
    met [tc_cases_left,tc_cases_right]
  ) THEN
  `finite_prefixes (hb Xh) Xh.acts` by ALL_TAC THEN1
  ( ps [finite_prefixes_def] THEN
    `!e. {e' | (e',e) IN hb Xh} SUBSET Xh.acts` by ALL_TAC THEN1
    ( fs ([SUBSET_DEF,LET_DEF] @ qsimps @ reloversimps) THEN
      met []
    ) THEN
    `FINITE Xh.acts` by met [lemHerdCondition_def] THEN
    met [lemHerdCondition_def,SUBSET_FINITE]
  ) THEN
  `w IN Xh.acts` by ALL_TAC THEN1
  ( fs reloversimps THEN met [] ) THEN
  `set_relation$acyclic (hb Xh)` by fs ([herdTheory.acyclic_def,hbAcyclic_def,set_relationTheory.acyclic_def,herdTheory.irreflexive_def,tch_def,lem_is_irreflexive_def] @ qsimps) THEN
  `FINITE {a | a <> r /\ is_write a /\ (loc_of a = loc_of r) /\ (a,r) IN hb Xh}` by ALL_TAC THEN1
  ( ps [finite_prefixes_def] THEN
    `{a | a <> r /\ is_write a /\ (loc_of a = loc_of r) /\ (a,r) IN hb Xh} SUBSET {e' | (e',r) IN hb Xh}` by fs [SUBSET_DEF] THEN
    met [SUBSET_FINITE]
  ) THEN
  `{a |
      a <> r /\ is_write a /\ (loc_of a = loc_of r) /\
        (a,r) IN hb Xh} <> {}` by ALL_TAC THEN1
  ( `w IN {a |
      a <> r /\ is_write a /\ (loc_of a = loc_of r) /\
      (a,r) IN hb Xh}` by ALL_TAC THEN1
    ( fs ([tch_def,hbAcyclic_def,herdTheory.acyclic_def,herdTheory.irreflexive_def,lem_is_irreflexive_def] @ qsimps) THEN
      met [tc_rules]
    ) THEN
    met [MEMBER_NOT_EMPTY]
  ) THEN
  `?x. x IN maximal_elements {a | a <> r /\ is_write a /\ (loc_of a = loc_of r) /\ (a,r) IN (hb Xh)} (hb Xh)` by met [finite_acyclic_has_maximal] THEN
  fs ([visible_side_effect_set_def,maximal_elements_def] @ qsimps) THEN
  met []
)









val lemHerdARF = store_thm("lemHerdARF",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  consistent_atomic_rf (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([herd_well_formed_rf_def,herd_provides_def,atomicRfConsistent_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,consistent_atomic_rf_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  Q.UNABBREV_TAC `Xw` THEN
  Q.UNABBREV_TAC `Xo` THEN
  fs ([hblem,doblem,swlem,LET_DEF,herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
  pats `irreflexive (hbarf Xh)` herdsimps THEN
  met []
)




val lemHerdCoherence = store_thm("lemHerdCoherence",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  coherent_memory_use (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([herd_well_formed_rf_def,well_formed_co_def,herd_provides_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,coherent_memory_use_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
      Q.UNABBREV_TAC `Xw` THEN
      Q.UNABBREV_TAC `Xo` THEN
      fs ([hblem,doblem,swlem,LET_DEF,herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
      pats `coherence Xh` ([coherence_def] @ herdsimps) THEN
      met []
)





val lemHerdRMW = store_thm("lemHerdRMW",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  rmw_atomicity (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([herd_well_formed_rf_def,well_formed_co_def,rf_for_every_read_def,herd_provides_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,rmw_atomicity_def,adjacent_less_than_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
      Q.UNABBREV_TAC `Xw` THEN
      Q.UNABBREV_TAC `Xo` THEN
      fs ([hblem,doblem,swlem,LET_DEF,herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
      rw [] THEN
      pats `rmwAtomicity Xh` ([rmwAtomicity_def] @ herdsimps) THEN
      `is_read b /\ is_write b` by ALL_TAC THEN1
      ( fs [is_read_def,is_write_def,is_RMW_def] THEN
        Cases_on `b` THEN
        fs []
      ) THEN
      `is_at_atomic_location Xh.lkh b` by met [rmwAtAtomicLoc,herd_well_formed_threads_def] THEN
      ps ([transitive_def,lem_is_strict_partial_order_def,lem_is_irreflexive_def] @ qsimps) THEN
      EQ_TAC THEN
      rw [] THENL
      [ ( `?c. (c,b) IN Xh.rfh` by met [] THEN
          `(c,b) IN Xh.co` by met [] THEN
          Cases_on `a = c` THEN1 met [] THEN
          `(a,c) IN Xh.co \/ (c,a) IN Xh.co` by met [sameLocThenAtomicLoc] THEN met []
        ),
        met [],
        met []
      ]
)





val lemHerdSCrf = store_thm("lemHerdSCrf",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  sc_accesses_sc_reads_restricted (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs ([herd_provides_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,sc_accesses_sc_reads_restricted_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  Q.UNABBREV_TAC `Xw` THEN
  Q.UNABBREV_TAC `Xo` THEN
  fs ([hblem,doblem,swlem,LET_DEF,herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF,herd_well_formed_rf_def] @ qsimps @ reloversimps) THEN
  rw [] THEN
  `is_acquire p_2` by met [scReadIsAcquire] THEN
  Cases_on `is_seq_cst p_1` THENL
  [ ( `is_release p_1` by met [scWriteIsRelease] THEN
      ps [] THEN
      `(p_1,p_2) IN hb Xh` by ALL_TAC THEN1
      ( Cases_on `tid_of p_1 = tid_of p_2` THEN1
        ( ps ([herd_well_formed_threads_def,well_formed_po_def] @ qsimps) THEN
          `(p_1,p_2) IN Xh.po` by ALL_TAC THEN1
          ( CCONTR_TAC THEN
            ps ([atomicRfConsistent_def] @ qsimps) THEN
            `p_2 <> p_1` by ALL_TAC THEN1
            ( CCONTR_TAC THEN
              ps ([well_formed_co_def] @ qsimps) THEN
              `is_RMW p_1` by met [readWriteIsRmw] THEN
              pats` rmwAtomicity Xh` ([rmwAtomicity_def] @ herdsimps) THEN
              met []
            ) THEN
            `(p_2,p_1) IN Xh.po` by met [] THEN
            `(p_2,p_1) IN hb Xh` by met [hb_def,IN_UNION] THEN
            pats `irreflexive (hbarf Xh)` (herdsimps @ [hbarf_def,arf_def]) THEN
            `is_at_atomic_location Xh.lkh p_2` by ALL_TAC THEN1
            ( `p_2 IN Xh.acts /\ is_read p_2` by met [] THEN
              pats `actions_respect_location_kinds Xh.acts Xh.lkh` ([actions_respect_location_kinds_def] @ qsimps) THEN
              specpat `Q` `p_2` THEN
              mppat `Q` THEN
              fs [is_at_atomic_location_def,is_read_def,is_seq_cst_def,loc_of_def] THEN
              Cases_on `p_2` THEN
              fs []
            ) THEN
            met []
          ) THEN
          ps [hb_def]
        ) THEN
        `(p_1,p_2) IN sw Xh` by ALL_TAC THEN1
        ( psg ([ext_def,seq_def,rcomp_def,A_def,stor_def,sw_def,swra_def]) THEN
          DISJ2_TAC THEN
          DISJ2_TAC THEN
          STRIP_TAC THEN1 met [] THEN
          Q.EXISTS_TAC `p_1` THEN
          psg [rs_def,seq_def,rcomp_def,crel_rw2,stor_def,A_def,W_def,rc_def,fid_def] THEN
          STRIP_TAC THEN1 met [relacq_writeread_atomic] THEN
          psg [rc_def,seq_def,rcomp_def,idf_def,cacq_rw2] THEN
          met [relacq_writeread_atomic]
        ) THEN
        psg [hb_def,ithb_def,ithbr_def,tch_def] THEN
        met [tc_rules,IN_UNION]
      ) THEN
      `(p_1,p_2) IN scp Xh` by ALL_TAC THEN1
      ( psg herdsimps THEN
        met []
      ) THEN
      `(p_1,p_2) IN sct` by met [tc_rules] THEN 
      CCONTR_TAC THEN ps [] THEN1 met [] THEN
      `(p_1,w') IN Xh.co` by ALL_TAC THEN1
      ( ps [lem_is_strict_total_order_on_def,lem_is_strict_partial_order_on_def,set_relationTheory.irreflexive_def] THEN
        `p_1 <> w'` by met [tc_rules] THEN
        ps [herd_well_formed_threads_def] THEN
        pats `well_formed_co Xh.acts Xh.lkh Xh.co` ([well_formed_co_def] @ qsimps) THEN
        `(p_1,w') IN Xh.co \/ (w',p_1) IN Xh.co` by met [SCAtAtomicLoc] THEN
        `(w',p_1) IN scp Xh` by ALL_TAC THEN1
        ( psg herdsimps THEN
          met []
        ) THEN
        ps ([lem_transitive_on_def] @ qsimps) THEN
        met [tc_rules]
      ) THEN
      `(p_2,w') IN scp Xh` by ALL_TAC THEN1
      ( ps [lem_is_strict_total_order_on_def,lem_is_strict_partial_order_on_def,set_relationTheory.irreflexive_def] THEN
        `p_2 <> w'` by met [tc_rules] THEN
        psg herdsimps THEN
        met []
      ) THEN
      `(p_2,w') IN sct` by met [tc_rules] THEN
      ps ([lem_is_strict_total_order_on_def,lem_is_strict_partial_order_on_def,lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
      met [tc_rules]
    ),
    ( ps [] THEN
      CCONTR_TAC THEN
      ps [] THEN
      `(p_2,w') IN scp Xh` by ALL_TAC THEN1
      ( psg herdsimps THEN
        `p_2 <> w'` by ALL_TAC THEN1
        ( fs ([lem_is_strict_partial_order_on_def,lem_is_strict_total_order_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
          met []
        ) THEN
        met []
      ) THEN
      ps ([lem_is_strict_partial_order_on_def,lem_is_strict_total_order_on_def,lem_transitive_on_def,set_relationTheory.irreflexive_def] @ qsimps) THEN
      met [tc_rules]
    )
  ]
)





val lemHerdTotEmpty = store_thm("lemHerdTotEmpty",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  tot_empty (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def] THEN
  psg [tot_empty_def] THEN
  Q.UNABBREV_TAC `Xw` THEN
  ps [herdToLemExecutionWitness_def,LET_DEF] THEN
  fs []
)










val lemHerdNARF = store_thm("lemHerdNARF",
``(
!Xh opsem herd_opsem p sct Xo Xw.
  lemHerdCondition herd_opsem opsem p /\
  herd_opsem p Xh /\
  cpp11$provides_clauses Xh /\
  tc (scp Xh) SUBSET sct /\
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  Abbrev (Xw = herdToLemExecutionWitness Xh sct) /\
  Abbrev (Xo = herdToLemPreExecution Xh) /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  opsem p Xo /\
  (hblem Xo Xw = hb Xh)
  ==>
  consistent_non_atomic_rf (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  fs [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def] THEN
  pats `nonAtomicRfConsistent Xh` [nonAtomicRfConsistent_def] THEN
  pats `Q`
  [is_empty_def,EXTENSION,FORALL_PROD,vse_def,seq_def,stor_def,rcomp_def,compr_def,hbl_def,narf_def,nonatomicloc_def,loc_def,W_def]
  THEN
  psg ([visible_side_effect_set_def,nonAtomicRfConsistent_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,consistent_non_atomic_rf_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  fs ([herd_well_formed_rf_def,herd_provides_def,visible_side_effect_set_def,nonAtomicRfConsistent_def,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,consistent_non_atomic_rf_def,LET_DEF,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  Q.UNABBREV_TAC `Xw` THEN
  Q.UNABBREV_TAC `Xo` THEN
  fs ([hblem,doblem,swlem,LET_DEF,herdToLemExecutionWitness_def,herdToLemPreExecution_def,LET_DEF] @ qsimps @ reloversimps) THEN
  met []
)






val lemHerdExecutionConsistent = store_thm("lemHerdExecutionConsistent",
``(
!Xh opsem herd_opsem p.
lemHerdCondition herd_opsem opsem p /\
herd_opsem p Xh /\
cpp11$provides_clauses Xh
==>
?sct.
  tc (scp Xh) SUBSET sct /\ 
  rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
  let Xo = herdToLemPreExecution Xh in
  let Xw = herdToLemExecutionWitness Xh sct in
  opsem p Xo /\
  (hblem Xo Xw = hb Xh) /\
  apply_tree with_consume_consistent_execution
       (Xo,Xw,with_consume_relations Xo Xw)
)``,
  rw [] THEN
  `?sct.
    tc (scp Xh) SUBSET sct /\ 
    rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
    lem_is_strict_total_order_on sct {a | a IN Xh.acts /\ is_seq_cst a} /\
    let Xo = herdToLemPreExecution Xh in
    let Xw = herdToLemExecutionWitness Xh sct in
    opsem p Xo /\
    (hblem Xo Xw = hb Xh)` by met [herdLemExistsEquivalentExecution] THEN
  Q.EXISTS_TAC `sct` THEN
  rw [] THEN fs [LET_DEF] THEN
  ps [with_consume_consistent_execution_def] THEN
  fs [apply_tree_rw,RES_FORALL,FORALL_PROD] THEN
  rw [] THENL
  [ ( MATCH_MP_TAC lemHerdAssumptions THEN met [] ),
    ( MATCH_MP_TAC lemHerdTotEmpty THEN met [] ),
    ( MATCH_MP_TAC lemHerdWFT THEN met [] ),
    ( MATCH_MP_TAC lemHerdWFrf THEN met [] ),
    ( MATCH_MP_TAC lemHerdConsistentLocks THEN met [] ),
    ( MATCH_MP_TAC lemHerdConsistentlo THEN met [] ),
    ( MATCH_MP_TAC lemHerdConsistentmo THEN met [] ),
    ( MATCH_MP_TAC lemHerdConsistentsc THEN met [] ),
    ( MATCH_MP_TAC lemHerdConsistentscFH THEN met [] ),
    ( MATCH_MP_TAC lemHerdConsistenthb THEN met [] ),
    ( MATCH_MP_TAC lemHerdDetRead THEN met [] ),
    ( MATCH_MP_TAC lemHerdNARF THEN met [] ),
    ( MATCH_MP_TAC lemHerdARF THEN met [] ),
    ( MATCH_MP_TAC lemHerdCoherence THEN met [] ),
    ( MATCH_MP_TAC lemHerdRMW THEN met [] ),
    ( MATCH_MP_TAC lemHerdSCrf THEN met [] )
  ]
)











val herdRaceThenLemRace = store_thm("herdRaceThenLemRace",
``(
!opsem herd_opsem p C.
  let Bh = calculate_program_behaviour model herd_opsem p in
  let Bl = behaviour with_consume_memory_model C opsem p in
  lemHerdCondition herd_opsem opsem p /\
  (Bh = Undef)
   ==>
  (Bl = Undefined)
)``,
  ps [calculate_program_behaviour_def,LET_DEF,RES_FORALL] THEN
  fs [model_def] THEN
  rw [] THEN
  ps [] THEN
  psg [behaviour_def,LET_DEF,RES_FORALL,true_condition_def,observable_filter_def,with_consume_memory_model_def] THEN
  rw [] THEN
  mppat `Q` THEN ps [] THEN rw [] THEN
  `?sct.
            tc (scp X) SUBSET sct /\
            rel_over sct {a | a IN X.acts /\ is_seq_cst a} /\
            lem_is_strict_total_order_on sct
              {a | a IN X.acts /\ is_seq_cst a} /\
            (let Xo = herdToLemPreExecution X in
             let Xw = herdToLemExecutionWitness X sct
             in
               opsem p Xo /\ (hblem Xo Xw = hb X) /\
               apply_tree with_consume_consistent_execution
                 (Xo,Xw,with_consume_relations Xo Xw))` by met [lemHerdExecutionConsistent] THEN
  fs [LET_DEF] THEN
  Q.EXISTS_TAC `(herdToLemPreExecution X,herdToLemExecutionWitness X sct,
         with_consume_relations (herdToLemPreExecution X)
           (herdToLemExecutionWitness X sct))` THEN
  fs [] THEN
  pats `Q = hb X` [hblem,doblem,swlem,LET_DEF,herdToLemExecutionWitness_def,herdToLemPreExecution_def] THEN
  psg ([hblem,doblem,swlem,LET_DEF,herdToLemExecutionWitness_def,herdToLemPreExecution_def,with_consume_relations_def] @ qsimps @ reloversimps) THEN
  fs [each_empty_def,locks_only_undefined_behaviour_def,LET_DEF] THEN
  ps [c11_base_intermediateTheory.requires_clauses_def,cpp11Theory.requires_clauses_def] THEN
  rw [] THENL
  [ ( DISJ2_TAC THEN DISJ1_TAC THEN
      ps [dataRace_def,data_races_def] THEN
      fs [] THEN
      pats `~is_empty (dr X)` [is_empty_def,dr_def,WM_def,MW_def,UNION_DEF,INTER_DEF,loc_def,cross_def,ext_def,inv_def,compr_def,comps_def,unis_def,na_def,W_def,M_def,lem_converse_def,SWAP_def,FST,SND,FORALL_PROD,A_def] THEN
      ps [EXTENSION] THEN
      rw [] THEN
      met [nonatomic_not_atomic,differentTidsDifferentActions,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,well_formed_threads_def,herd_well_formed_threads_def,A_def]
    ),
    ( DISJ1_TAC THEN
      `well_formed_po X.acts X.po` by ps [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,herd_well_formed_threads_def] THEN
      ps [unsequencedRace_def] THEN
      pats `~is_empty (ur X)` [is_empty_def,ur_def,WM_def,MW_def,UNION_DEF,INTER_DEF,loc_def,cross_def,ext_def,inv_def,compr_def,comps_def,unis_def,na_def,W_def,M_def,lem_converse_def,SWAP_def,FST,SND,FORALL_PROD,stor_def,int_def] THEN
      rw [] THEN
      ps [EXTENSION,well_formed_po_def,RES_FORALL,tch_def] THEN
      rw [] THEN
      met [tc_rules]
    ),
    ( DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      ps [badLock_def,locks_only_bad_mutexes_def,locks_only_good_mutex_use_def,RES_FORALL,RES_EXISTS] THEN
      pats `~is_empty (bl X)` [is_empty_def,bl_def,UNION_DEF,INTER_DEF,loc_def,cross_def,ext_def,inv_def,compr_def,comps_def,unis_def,lem_converse_def,SWAP_def,FST,SND,FORALL_PROD,seq_def,stor_def,rcomp_def,lk_def,ls_def,ul_def] THEN
      rw [] THEN
      ps [EXTENSION] THEN
      rw [] THEN
      met []
    ),
    ( DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      `well_formed_lo X.acts X.lkh X.loh` by ps [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def] THEN
      ps [badUnlock_def,locks_only_bad_mutexes_def,locks_only_good_mutex_use_def,RES_FORALL,RES_EXISTS] THEN
      pats `~is_empty (lu X)` [is_empty_def,lu_def,UNION_DEF,INTER_DEF,loc_def,cross_def,ext_def,inv_def,compr_def,comps_def,unis_def,lem_converse_def,SWAP_def,FST,SND,FORALL_PROD,seq_def,stor_def,rcomp_def,lk_def,ls_def,ul_def,lopowoul_def] THEN
      rw [] THEN
      ps [EXTENSION] THEN
      rw [] THEN
      ps [well_formed_lo_def,rel_over_def,domain_def,range_def,SUBSET_DEF] THEN
      met []
    )
  ]
)








val tcsbeqsb = store_thm("tcsbeqsb",
``(
!opsem herd_opsem p Xo.
  lemHerdCondition herd_opsem opsem p /\ opsem p Xo ==> (tc Xo.sb = Xo.sb)
)``,
  rw [] THEN
  psg [EXTENSION,FORALL_PROD] THEN
  rw [] THEN
  EQ_TAC THENL
  [ ( rw [] THEN
      `well_formed_po Xo.actions Xo.sb` by
      ( fs [lemHerdCondition_def] THEN
        `?data addr. herd_opsem p (lemToHerdExecution Xo Xw addr data)` by met [] THEN
        `well_formed_po (lemToHerdExecution Xo Xw addr data).acts (lemToHerdExecution Xo Xw addr data).po` by met [] THEN
        pats `Q` [lemToHerdExecution_def,LET_DEF] THEN
        met []
      ) THEN
      ps [well_formed_po_def,lem_is_strict_partial_order_def] THEN
      met [transitive_tc]
    ),
    met [tc_rules]
  ]
)




val lemRaceThenHerdRace = store_thm("lemRaceThenHerdRace",
``(
!opsem herd_opsem p.
  let Bh = calculate_program_behaviour model herd_opsem p in
  let Bl = behaviour with_consume_memory_model true_condition opsem p in
  lemHerdCondition herd_opsem opsem p /\
  (Bl = Undefined)
   ==>
  (Bh = Undef)
)``,
  rw [] THEN
  ps [behaviour_def,LET_DEF,RES_FORALL,true_condition_def,observable_filter_def,with_consume_memory_model_def] THEN
  rw [] THEN
  fs [] THEN
  Cases_on `!X.
           (?Xo Xw.
              (X = (Xo,Xw,with_consume_relations Xo Xw)) /\
              opsem p Xo /\
              apply_tree with_consume_consistent_execution
                (Xo,Xw,with_consume_relations Xo Xw)) ==>
           each_empty locks_only_undefined_behaviour X` THEN1
  ( mppat `Q = Undefined` THEN
    rw [] THEN
    met []
  ) THEN
  mppat `Q = Undefined` THEN
  rw [] THEN
  Q.UNABBREV_TAC `Bh` THEN
  psg [calculate_program_behaviour_def,LET_DEF,RES_FORALL] THEN
  fs [model_def] THEN
  rw [] THEN
  mppat `Q` THEN ps [] THEN rw [] THEN
  `?data addr.
       (let Xh = lemToHerdExecution Xo Xw addr data
        in
          herd_opsem p Xh /\ (tc (data UNION addr) = Xo.dd) /\
          (hblem Xo Xw = hb Xh) /\ cpp11$provides_clauses Xh)` by met [lemHerdExistsConsistentExecution] THEN
  ps [LET_DEF] THEN
  Q.EXISTS_TAC `(lemToHerdExecution Xo Xw addr data)` THEN
  fs [] THEN
  ps [with_consume_relations_def,LET_DEF] THEN    
  Q.ABBREV_TAC `hrs =
                  hypothetical_release_sequence_set Xo.actions Xo.lk
                    Xw.mo` THEN
  Q.ABBREV_TAC `rs' = release_sequence_set Xo.actions Xo.lk Xw.mo` THEN
  Q.ABBREV_TAC `sw' =
                  release_acquire_fenced_synchronizes_with_set
                    Xo.actions Xo.sb Xo.asw Xw.rf Xw.lo rs' hrs
            ` THEN
  Q.ABBREV_TAC `cad' = with_consume_cad_set Xo.actions Xo.sb Xo.dd Xw.rf` THEN
  Q.ABBREV_TAC `dob' = with_consume_dob_set Xo.actions Xw.rf rs' cad'` THEN
  Q.ABBREV_TAC `ithb' =
                  inter_thread_happens_before Xo.actions Xo.sb sw' dob'` THEN
  Q.ABBREV_TAC `hb' = happens_before Xo.actions Xo.sb ithb'` THEN
  Q.ABBREV_TAC `vse' = visible_side_effect_set Xo.actions hb'` THEN
  fs [each_empty_def,locks_only_undefined_behaviour_def,LET_DEF] THEN
  ps [c11_base_intermediateTheory.requires_clauses_def,cpp11Theory.requires_clauses_def] THENL
  [ ( DISJ2_TAC THEN DISJ1_TAC THEN
      pats `unsequenced_races Q <> {}` [unsequenced_races_def] THEN
      ps [tch_def,unsequencedRace_def,is_empty_def,ur_def,WM_def,MW_def,UNION_DEF,INTER_DEF,loc_def,cross_def,ext_def,inv_def,compr_def,comps_def,unis_def,na_def,W_def,M_def,lem_converse_def,SWAP_def,FST,SND,FORALL_PROD,stor_def,int_def] THEN
      fs [lemToHerdExecution_def,hblem,LET_DEF] THEN
      rw [] THEN
      psg [EXTENSION,FORALL_PROD,EXISTS_PROD] THEN
      pats `Q` [EXTENSION] THEN
      rw [] THEN
      met [tcsbeqsb,sameLocThenNALoc,NALocThenReadOrWrite]
    ),
    ( DISJ1_TAC THEN
      ps [dataRace_def,is_empty_def,dr_def,WM_def,MW_def,UNION_DEF,INTER_DEF,loc_def,cross_def,ext_def,inv_def,compr_def,comps_def,unis_def,na_def,W_def,M_def,lem_converse_def,SWAP_def,FST,SND,FORALL_PROD,EXISTS_PROD,A_def] THEN
      pats `data_races Q <> {}` [data_races_def] THEN
      fs [lemToHerdExecution_def,LET_DEF] THEN
      pats `Q` [EXTENSION,FORALL_PROD,EXISTS_PROD] THEN
      psg [EXTENSION,FORALL_PROD,EXISTS_PROD] THEN
      ps [hblem,swlem,doblem,LET_DEF] THEN
      met [lem_consistent_then_well_formed_actions]
    ),
    ( fs [lemToHerdExecution_def,hblem,doblem,swlem,LET_DEF,cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,herd_well_formed_threads_def] THEN
      pats `well_formed_inits Q Q' Q''` [well_formed_inits_def,RES_FORALL,RES_EXISTS,M_def,SUBSET_DEF,EXTENSION] THEN
      pats `well_formed_asw Q Q' Q''` [well_formed_inits_def,RES_FORALL,RES_EXISTS,M_def,SUBSET_DEF,EXTENSION] THEN      
      pats `indeterminate_reads Q <> {}` [indeterminate_reads_def,RES_EXISTS,EXTENSION] THEN
      ps [] THEN
      met []
    ),
    ( DISJ2_TAC THEN
      DISJ2_TAC THEN
      fs [lemToHerdExecution_def,hblem,LET_DEF] THEN
      pats `locks_only_bad_mutexes Q <> {}` [locks_only_bad_mutexes_def,with_consume_relations_def,locks_only_good_mutex_use_def,RES_FORALL,RES_EXISTS,EXTENSION] THEN
      ps [] THENL
      [ ( DISJ2_TAC THEN
          psg [badUnlock_def,badLock_def,is_empty_def,lu_def,UNION_DEF,INTER_DEF,loc_def,cross_def,ext_def,inv_def,compr_def,comps_def,unis_def,lem_converse_def,SWAP_def,FST,SND,FORALL_PROD,seq_def,stor_def,rcomp_def,lk_def,ls_def,ul_def,lopowoul_def,EXTENSION] THEN
          fs [] THEN
          ps [with_consume_consistent_execution_def] THEN
          fs [apply_tree_rw,locks_only_consistent_lo_def,relation_over_def,rel_over_def,domain_def,range_def,SUBSET_DEF] THEN
          met []
        ),
        ( DISJ1_TAC THEN
          psg [badUnlock_def,badLock_def,is_empty_def,bl_def,lu_def,UNION_DEF,INTER_DEF,loc_def,cross_def,ext_def,inv_def,compr_def,comps_def,unis_def,lem_converse_def,SWAP_def,FST,SND,FORALL_PROD,seq_def,stor_def,rcomp_def,lk_def,ls_def,ul_def,lopowoul_def,EXTENSION] THEN
          ps [with_consume_consistent_execution_def] THEN
          fs [apply_tree_rw,locks_only_consistent_lo_def,relation_over_def,rel_over_def,domain_def,range_def,SUBSET_DEF] THEN
          met []
        )
      ]
    )
  ]
)











val lemHerdEquiRacy = store_thm("lemHerdEquiRacy",
``(
!opsem herd_opsem p.
  let Bh = calculate_program_behaviour model herd_opsem p in
  let Bl = behaviour with_consume_memory_model true_condition opsem p in
  lemHerdCondition herd_opsem opsem p ==>
  ((Bh = Undef) = (Bl = Undefined))
)``,
  rw [] THEN
  EQ_TAC THEN
  rw [] THENL
  [ met [herdRaceThenLemRace],
    met [lemRaceThenHerdRace]
  ]
)





val lemHerdEquivalent = store_thm("lemHerdEquivalent",
``(! herd_opsem opsem p.
  lemHerdCondition herd_opsem opsem p ==>
  lemHerdBehaviourEquivalent with_consume_memory_model model herd_opsem opsem p
)``,
  ps [lemHerdBehaviourEquivalent_def] THEN
  rw [] THEN
  `(Bh = Undef) = (Bl = Undefined)` by met [lemHerdEquiRacy] THEN
  Cases_on `(Bh = Undef) /\ (Bl = Undefined)` THEN1 met [] THEN
  `(Bh <> Undef) /\ (Bl <> Undefined)` by met [] THEN
  fs [] THEN rw [] THEN
  Q.UNABBREV_TAC `Bh` THEN
  Q.UNABBREV_TAC `Bl` THEN
  fs ([calculate_program_behaviour_def,behaviour_def,LET_DEF,true_condition_def] @ qsimps) THEN
  Q.EXISTS_TAC `{X | herd_opsem p X /\ model.provides X}` THEN
  Q.EXISTS_TAC `(observable_filter
           {(Xo,Xw,rl) |
            opsem p Xo /\
            apply_tree with_consume_memory_model.consistent
              (Xo,Xw,rl) /\
            (rl =
             with_consume_memory_model.relation_calculation Xo Xw)})` THEN
  rw [] THEN1 met [] THEN1 met [] THEN
  rmpat `Q` THEN
  rmpat `Q` THEN  
  ps [with_consume_memory_model_def,model_def,lemHerdExecutionSetsEquivalent_def,RES_FORALL,RES_EXISTS,FORALL_PROD,observable_filter_def] THEN
  rw [] THENL
  [ ( `?data addr.
       (let Xh = lemToHerdExecution p_1 p_2 addr data
        in
          herd_opsem p Xh /\ (tc (data UNION addr) = p_1.dd) /\
          (hblem p_1 p_2 = hb Xh) /\
          scp (lemToHerdExecution p_1 p_2 addr data) SUBSET p_2.sc /\
          cpp11$provides_clauses Xh)` by met [lemHerdExistsConsistentExecution] THEN
      fs [LET_DEF] THEN
      Q.EXISTS_TAC `(lemToHerdExecution p_1 p_2 addr data)` THEN
      ps [] THEN
      psg ([lemHerdPreExecutionEquivalent_def,lemHerdExecutionEquivalent_def,lemToHerdExecution_def,lemHerdPartialExecutionEquivalent_def,LET_DEF]) THEN
      fs [] THEN
      rw [] THENL
      [ ( Q.ABBREV_TAC `Xh = lemToHerdExecution p_1 p_2 addr data` THEN
           `well_formed_inits Xh.inits Xh.acts Xh.po /\
            well_formed_asw Xh.acts Xh.inits Xh.aswh` by ps [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,herd_well_formed_threads_def] THEN
           Q.UNABBREV_TAC `Xh` THEN
           fs ([SUBSET_DEF,well_formed_inits_def,well_formed_asw_def,lemToHerdExecution_def,LET_DEF] @ qsimps) THEN
           ps (herdsimps @ qsimps) THEN fs [] THEN psg [EXTENSION] THEN rw [] THEN
           Cases_on `x` THEN
           fs [] THEN
           met []
         ),
         ( ps (herdsimps @ qsimps) THEN fs [] THEN psg [EXTENSION] THEN rw []
         ),
         ( pats `Q SUBSET Q'` ([lemToHerdExecution_def,LET_DEF] @ qsimps) THEN
           met []
         )
      ]
    ),
    ( `?sct.
       tc (scp Xh) SUBSET sct /\
       rel_over sct {a | a IN Xh.acts /\ is_seq_cst a} /\
       lem_is_strict_total_order_on sct
         {a | a IN Xh.acts /\ is_seq_cst a} /\
       (let Xo = herdToLemPreExecution Xh in
        let Xw = herdToLemExecutionWitness Xh sct
        in
          opsem p Xo /\ (hblem Xo Xw = hb Xh) /\
          apply_tree with_consume_consistent_execution
            (Xo,Xw,with_consume_relations Xo Xw))` by met [lemHerdExecutionConsistent] THEN
      fs [LET_DEF] THEN
      Q.EXISTS_TAC `(herdToLemPreExecution Xh,herdToLemExecutionWitness Xh sct)` THEN
      fs [] THEN
      psg ([lemHerdPreExecutionEquivalent_def,lemHerdExecutionEquivalent_def,herdToLemPreExecution_def,herdToLemExecutionWitness_def,lemHerdPartialExecutionEquivalent_def,LET_DEF]) THEN
      fs [] THEN
      rw [] THENL
      [ ( `well_formed_inits Xh.inits Xh.acts Xh.po /\
            well_formed_asw Xh.acts Xh.inits Xh.aswh` by ps [cpp11Theory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def,herd_provides_def,herd_well_formed_threads_def] THEN
          fs ([SUBSET_DEF,well_formed_inits_def,well_formed_asw_def,lemToHerdExecution_def,LET_DEF] @ qsimps) THEN
          ps (herdsimps @ qsimps) THEN fs [] THEN psg [EXTENSION] THEN rw [] THEN
          fs [] THEN
          met []
        ), 
        ( fs ([SUBSET_DEF] @ qsimps) THEN
          met [tc_rules]
        )
      ]
    )
  ]
)


				 
val _ = export_theory()
