open HolKernel boolLib Parse bossLib stringTheory pred_setTheory pairTheory res_quanTheory PairRules optionTheory relationTheory quantHeuristicsLib pairSimps pred_setSimps proofManagerLib


open numTheory
open arithmeticTheory


open Defn
open listTheory

open c11Theory;
open c11_base_intermediateTheory;


open cpp11Theory;
open cpp11soundnessTheory;

open herdTheory;

open lem_relationTheory


open set_relationTheory;

open c11_baseTheory;
open cpp11_paperTheory;
open cpp11soundness_paperTheory;

open metaTheory;


val _ = new_theory "meta_paper"

fun fs thms = FULL_SIMP_TAC (srw_ss()) thms;
fun gs thms = FULL_SIMP_TAC (std_ss ++ gen_beta_ss) thms;
fun stdfs thms = FULL_SIMP_TAC (std_ss) thms;
fun stds thms = SIMP_TAC (std_ss) thms;
fun ps thms = FULL_SIMP_TAC (std_ss ++ PRED_SET_ss) thms;
fun psg thms = SIMP_TAC (std_ss ++ PRED_SET_ss) thms;
fun rw thms = RW_TAC (srw_ss()) thms;
val st      = REPEAT STRIP_TAC;
val dg      = REPEAT (DISCH_TAC ORELSE GEN_TAC);
fun show1 term tactic = term by ALL_TAC THEN1 tactic;
fun ts thms = FULL_SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) thms;
fun simp_ts thms = SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) thms;
fun asm_ts thms = ASM_SIMP_TAC (std_ss ++ QUANT_INST_ss [std_qp]) thms;

fun mppat quote = Q.PAT_ASSUM quote MP_TAC;
fun rmpat quote = Q.PAT_ASSUM quote (fn x => ALL_TAC);
fun specpat quote specq = Q.PAT_ASSUM quote (fn x => ASSUME_TAC (Q.SPEC specq x));
fun specpatadd quote specq = Q.PAT_ASSUM quote
  (fn x => ASSUME_TAC (x) THEN ASSUME_TAC (Q.SPEC specq x));
fun mpspecpat quote specq = Q.PAT_ASSUM quote (fn x => MP_TAC (Q.SPEC specq x));

fun homatchmppat quote = Q.PAT_ASSUM quote HO_MATCH_MP_TAC;
fun matchmppat quote = Q.PAT_ASSUM quote MATCH_MP_TAC;

fun specpatp quote q1 q2 = Q.PAT_ASSUM quote
                           (fn x => ASSUME_TAC (Q.SPEC q2 (Q.SPEC q1 x)));
fun gs thms = SIMP_TAC (srw_ss()) thms;
fun met thms = METIS_TAC thms;
fun pats quote thms = Q.PAT_ASSUM quote (fn x => ASSUME_TAC (SIMP_RULE (srw_ss()) thms x));

fun impresp quote = Q.PAT_ASSUM quote IMP_RES_TAC;

fun popspec quote = POP_ASSUM (fn x => ASSUME_TAC (Q.SPEC quote x));

fun popsimp thms = POP_ASSUM (fn x => ASSUME_TAC (SIMP_RULE (srw_ss()) thms x));


val reloversimps' =
  [ rel_over_def,RES_FORALL,relation_over_def,range_def,domain_def,SUBSET_DEF ]

    
val both_coh_def = Define `both_coh X =
		     herd_provides X /\
( ( Coh X /\ Hb X
  ) \/
  ( coherence X /\ hbAcyclic X
  )
)`

val hbo_def = Define `hbo X = c11_base_intermediate$hb X`
val hbp_def = Define `hbp X = c11_base$hb X`


val po_in_both_hb = store_thm("po_in_both_hb",
``( !X a b. (a,b) IN X.po ==>
       (a,b) IN hbp X /\
       (a,b) IN hbo X
  )``,
  rw [] THEN
  ps [      hbo_def, hbp_def,
            c11_baseTheory.hb_def,
	    c11_base_intermediateTheory.hb_def,
	    c11_baseTheory.sb_def
     ]
)

val both_coh_co_over_acts = store_thm("both_coh_co_over_acts",
``( !X a b. both_coh X /\ (a,b) IN X.co ==> a IN X.acts /\ b IN X.acts /\ (loc_of a = loc_of b) /\ (a <> b) /\ is_write a /\ is_write b /\ a IN atomicloc X /\ b IN atomicloc X /\ (loc_of a = loc_of b)
  )``,
  rw [] THEN
  ps [both_coh_def,herd_provides_def,well_formed_co_def,atomicloc_def,loc_def] THEN
  ps reloversimps' THEN
  met []
)

val both_then_either_co = store_thm("both_then_either_co",
``( !X. both_coh X ==> Coh X \/ coherence X
  )``,
  ps [both_coh_def] THEN
  met []
)

  
val both_then_coww = store_thm("both_then_coww",
``( !X. both_coh X ==>
        (~?a b. (a,b) IN hbp X /\ (b,a) IN X.co) \/
        (~?a b. (a,b) IN hbo X /\ (b,a) IN X.co)
  )``,
  rw [] THEN
  CCONTR_TAC THEN
  `Coh X \/ coherence X` by met [both_then_either_co] THEN  
  ps [c11_base_intermediateTheory.coherence_def,c11_baseTheory.Coh_def,c11_baseTheory.mo_def,herdTheory.irreflexive_def,seq_def,rc_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD,rcomp_def,coww_def,c11_baseTheory.hbl_def,c11_base_intermediateTheory.hbl_def,inv_def,lem_converse_def,SWAP_def] THEN
  mppat `!p. P` THEN
  ps [] THENL
  [ ( Q.EXISTS_TAC `b` THEN
      Q.EXISTS_TAC `b` THEN
      STRIP_TAC THEN1 met [both_coh_co_over_acts] THEN
      Q.EXISTS_TAC `a`
    ),
    ( Q.EXISTS_TAC `b'` THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      DISJ2_TAC THEN
      Q.EXISTS_TAC `a'`
    )
    ] THEN
  ps [loc_def,hbo_def,hbp_def] THEN
  met [both_coh_co_over_acts]
)	      


val int_neq_then_po = store_thm("int_neq_then_po",
``( !X a b. both_coh X ==>
      (tid_of a = tid_of b) /\ (a <> b) /\  a IN X.acts /\ b IN X.acts ==>
      (a,b) IN X.po \/ (b,a) IN X.po
  )``,
  rw [] THEN
  ps [both_coh_def,herd_provides_def,well_formed_po_def,herd_well_formed_threads_def]
)



val both_coh_atom_co = store_thm("both_coh_atom_co",
``( !X a b. both_coh X /\ a IN X.acts /\ b IN X.acts /\ (loc_of a = loc_of b) /\ (a <> b) /\ is_write a /\ is_write b /\ a IN atomicloc X ==> (a,b) IN X.co \/ (b,a) IN X.co
  )``,
  rw [] THEN
  ps [both_coh_def,herd_provides_def,well_formed_co_def,atomicloc_def] THEN
  ps reloversimps' THEN
  met []
)

val int_po_then_stuff = store_thm("int_po_then_stuff",
``( !X a b. both_coh X /\ (a,b) IN X.po 
      ==>
      (tid_of a = tid_of b) /\ (a <> b) /\  a IN X.acts /\ b IN X.acts
  )``,
  rw [] THEN
  ps ([both_coh_def,herd_provides_def,well_formed_po_def,herd_well_formed_threads_def,RES_FORALL,FORALL_PROD] @ reloversimps') THEN
  met []  
)



val rf_then_stuff = store_thm("rf_then_stuff",
``( !X a b. both_coh X /\ (a,b) IN X.rfh
      ==>
      (loc_of a = loc_of b) /\ a IN X.acts /\ b IN X.acts /\ a IN W X /\ b IN R X
  )``,
  rw [] THEN
  ps ([both_coh_def,herd_provides_def,herd_well_formed_rf_def,herd_well_formed_threads_def,RES_FORALL,FORALL_PROD,W_def,R_def] @ reloversimps') THEN
  met []  
)

				 

val coi_poW_eq = store_thm("coi_poW_eq",
``( !X. both_coh X ==>
   (coi X = X.po INTER loc X INTER WW X INTER (cross (atomicloc X) (atomicloc X)))
  )``,
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  STRIP_TAC THEN
  ps [coi_def,int_def,WW_def,cross_def,W_def,stor_def,loc_def] THEN
  `(q,r) IN X.po \/ (r,q) IN X.po` by met [both_coh_co_over_acts,int_neq_then_po] THEN
  met [both_coh_co_over_acts,po_in_both_hb,both_then_coww,both_coh_atom_co,int_po_then_stuff,po_in_both_hb]
)
  
  

val R_W_rmw = store_thm("R_W_rmw",
``( !X a.
   (a IN R X /\ a IN W X ==> a IN rmw X) /\
   (a IN rmw X ==> a IN R X /\ a IN W X)
  )``,
  rw [] THEN
  ps [rmw_def,R_def,W_def,is_read_def,is_write_def,is_RMW_def] THEN
  Cases_on `a` THEN  
  fs []
)  




val swraEquivalent = store_thm("swraEquivalent",
``( !X. both_coh X ==>
   (c11_base$swra X = c11_base_intermediate$swra X)
  )``,  
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  rw [] THEN
  ps [c11_baseTheory.swra_def,
      c11_base_intermediateTheory.swra_def,
      c11_baseTheory.fsb_def,
      c11_baseTheory.sbf_def,
      c11_base_intermediateTheory.fid_def,
      c11_base_intermediateTheory.idf_def,
      ext_def,seq_def,rcomp_def,
      c11_base_intermediateTheory.rs_def,
      c11_baseTheory.sb_def,
      compr_def,comps_def,stor_def,
      c11_baseTheory.crel_def,
      c11_base_intermediateTheory.crel_def
     ] THENL
  [ Q.EXISTS_TAC `z'`,
    Q.EXISTS_TAC `z'`,
    Q.EXISTS_TAC `z'`,
    Q.EXISTS_TAC `z'`,
    Q.EXISTS_TAC `z'`,
    Q.EXISTS_TAC `z'`,
    Q.EXISTS_TAC `z'`,
    Q.EXISTS_TAC `z'` ] THEN
  rw [] THENL
  [ Q.EXISTS_TAC `z`, Q.EXISTS_TAC `z''`,
    Q.EXISTS_TAC `z`, Q.EXISTS_TAC `z''`,
    Q.EXISTS_TAC `z`, Q.EXISTS_TAC `z''`,
    Q.EXISTS_TAC `z`, Q.EXISTS_TAC `z''`,
    Q.EXISTS_TAC `z`,
    Q.EXISTS_TAC `z`,
    Q.EXISTS_TAC `z`,
    Q.EXISTS_TAC `z` ] THEN
  ps [rc_def,c11_baseTheory.rs_def,c11_baseTheory.rs_prime_def,c11_baseTheory.mo_def,coi_def,coe_def,int_def,ext_def,cross_def,unis_def,DIFF_DEF,seq_def,rcomp_def,c11_baseTheory.cacq_def,c11_base_intermediateTheory.cacq_def] THEN
  `z' IN X.acts` by ps [R_def] THEN  
  met [R_W_rmw,both_coh_co_over_acts,rf_then_stuff]
)


val swulEquivalent = store_thm("swulEquivalent",
``( !X. both_coh X ==>
   (c11_base$swul X = c11_base_intermediate$swul X)
  )``,  
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  rw [] THEN
  ps [ c11_baseTheory.swul_def,
       c11_base_intermediateTheory.swul_def ]
)



			      
val aswEquivalent = store_thm("aswEquivalent",
``( !X. both_coh X ==>
   (c11_base$pp_asw X = c11_base_intermediate$pp_asw X)
  )``,  
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  rw [] THEN
  ps [ c11_baseTheory.pp_asw_def,
       c11_base_intermediateTheory.pp_asw_def,
       c11_baseTheory.sb_def
     ]
)



val swEquivalent = store_thm("swEquivalent",
``( !X. both_coh X ==>
   (c11_base$sw X = c11_base_intermediate$sw X)
  )``,  
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  rw [] THEN
  ps [ c11_baseTheory.sw_def,
       c11_base_intermediateTheory.sw_def
     ] THEN
  met [aswEquivalent,swulEquivalent,swraEquivalent]
)

val ddEquivalent = store_thm("ddEquivalent",
``( !X. both_coh X ==>
   (c11_base$dd X = c11_base_intermediate$dd X)
  )``,  
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  rw [] THEN
  ps [ c11_baseTheory.dd_def,
       c11_base_intermediateTheory.dd_def,
       c11_baseTheory.sb_def
     ] THEN
  met []
)


val cadEquivalent = store_thm("cadEquivalent",
``( !X. both_coh X ==>
   (c11_base$cad X = c11_base_intermediate$cad X)
  )``,  
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  rw [] THEN
  ps [ c11_baseTheory.cad_def,
       c11_base_intermediateTheory.cad_def,
       c11_baseTheory.sb_def
     ] THEN
  met [ddEquivalent]
)

val fullp_rs_def = Define `fullp_rs X =
      seq (stor (c11_base$crel X))
          (seq (rc X (c11_base$fsb X))
               (seq (stor (A X INTER W X))
                    (rc X (c11_base$rs X))))`



val rsEquivalent = store_thm("rsEquivalent",
``( !X. both_coh X ==>
   (fullp_rs X = c11_base_intermediate$rs X)
  )``,  
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  rw [] THEN
  ps [ fullp_rs_def,
      c11_baseTheory.fsb_def,
      c11_baseTheory.sbf_def,
      c11_base_intermediateTheory.fid_def,
      c11_base_intermediateTheory.idf_def,
      ext_def,seq_def,rcomp_def,
      c11_base_intermediateTheory.rs_def,
      c11_baseTheory.sb_def,
      compr_def,comps_def,stor_def,
      c11_baseTheory.crel_def,
      c11_base_intermediateTheory.crel_def
     ] THEN
  Q.EXISTS_TAC `z` THEN
  rw [] THEN
  ps [rc_def,c11_baseTheory.rs_def,c11_baseTheory.rs_prime_def,c11_baseTheory.mo_def,coi_def,coe_def,int_def,ext_def,cross_def,unis_def,DIFF_DEF,seq_def,rcomp_def,c11_baseTheory.cacq_def,c11_base_intermediateTheory.cacq_def] THEN
  met [R_W_rmw,both_coh_co_over_acts,rf_then_stuff]
)
			    			     

val dobEquivalent = store_thm("dobEquivalent",
``( !X. both_coh X ==>
   (c11_base$dob X = c11_base_intermediate$dob X)
  )``,  
  ps [EXTENSION] THEN
  rw [] THEN Cases_on `x` THEN
  EQ_TAC THEN
  rw [] THEN
  `fullp_rs X = c11_base_intermediate$rs X` by met [rsEquivalent] THEN
  ps [ c11_baseTheory.dob_def,
       c11_base_intermediateTheory.dob_def,
       c11_baseTheory.sb_def,
       fullp_rs_def
     ] THEN
  ps [EXTENSION,FORALL_PROD] THEN  
  ps [seq_def,ext_def,rcomp_def,stor_def] THEN
  Q.EXISTS_TAC `z` THEN
  ps [c11_baseTheory.ccon_def,
      c11_base_intermediateTheory.ccon_def,
      c11_baseTheory.crel_def,
      c11_base_intermediateTheory.crel_def] THEN
  rw [] THEN
  met [cadEquivalent]			    
)			    
     
val hbEquivalent = store_thm("hbEquivalent",
``(  !X. both_coh X ==> (c11_base$hb X = c11_base_intermediate$hb X)
  )``,
  ps [FUN_EQ_THM,IN_DEF,
      c11_baseTheory.hb_def,
      c11_base_intermediateTheory.hb_def,
      c11_baseTheory.sb_def,
      c11_baseTheory.ithb_def,
      c11_base_intermediateTheory.ithb_def,
      c11_baseTheory.ithbr_def,
      c11_base_intermediateTheory.ithbr_def
     ] THEN
  rw [] THEN
  EQ_TAC THEN
  rw [] THEN
  met [dobEquivalent,swEquivalent]
)  
  


val coh_then_paper_coh = store_thm("coh_then_paper_coh",
``( !X. (c11_base$hb X = c11_base_intermediate$hb X) /\ herd_provides X /\
      c11_base_intermediate$coherence X /\ c11_base_intermediate$hbAcyclic X ==>
      c11_base$Coh X
  )``,
  rw [] THEN
  `both_coh X` by ps [both_coh_def] THEN
  ps [c11_base_intermediateTheory.coherence_def,
      c11_base_intermediateTheory.corr_def,
      c11_base_intermediateTheory.corw_def,
      c11_base_intermediateTheory.cowr_def,
      c11_base_intermediateTheory.coww_def,
      c11_baseTheory.Coh_def,
     herdTheory.irreflexive_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD] THEN
  rw [] THEN
  specpat `!p. P` `p_1` THEN
  ps [rc_def,seq_def,rcomp_def,fr_def,c11_base_intermediateTheory.hbl_def,inv_def,lem_converse_def,FORALL_PROD,SWAP_def,c11_baseTheory.mo_def,loc_def,herdTheory.fr_def] THEN
  rw [] THEN
  CCONTR_TAC THEN
  ps [] THEN
  met [rf_then_stuff,both_coh_co_over_acts]
)



val paper_coh_then_coh = store_thm("paper_coh_then_coh",
``( !X. (c11_base$hb X = c11_base_intermediate$hb X) /\ herd_provides X /\
      c11_base$Coh X /\ c11_base$Hb X ==>
      c11_base_intermediate$coherence X
  )``,
  rw [] THEN
  `both_coh X` by ps [both_coh_def] THEN
  ps [c11_base_intermediateTheory.coherence_def,
      c11_base_intermediateTheory.corr_def,
      c11_base_intermediateTheory.corw_def,
      c11_base_intermediateTheory.cowr_def,
      c11_base_intermediateTheory.coww_def,
      c11_baseTheory.Coh_def,
     herdTheory.irreflexive_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD] THEN
  rw [] THEN
  specpat `!p. P` `p_1` THEN
  ps [rc_def,seq_def,rcomp_def,herdTheory.fr_def,c11_base_intermediateTheory.hbl_def,herdTheory.inv_def,lem_converse_def,FORALL_PROD,SWAP_def,c11_baseTheory.mo_def,loc_def] THEN
  rw [] THEN
  CCONTR_TAC THEN
  ps [] THEN
  met [rf_then_stuff,both_coh_co_over_acts]
)



val rf_then_loc = store_thm("rf_then_loc",
``( !X a b. herd_provides X /\
      (a,b) IN X.rfh ==>
      (a IN atomicloc X /\ b IN atomicloc X) \/
      (a IN nonatomicloc X /\ b IN nonatomicloc X)
  )``,
  rw [] THEN
  `is_read b /\ is_write a /\ (loc_of a = loc_of b) /\ a IN X.acts /\ b IN X.acts` by ALL_TAC THEN1
  ( ps [herd_provides_def,herd_well_formed_rf_def,RES_FORALL,FORALL_PROD] THEN
    met []
  ) THEN
  `actions_respect_location_kinds X.acts X.lkh` by ALL_TAC THEN1
  ( ps [herd_provides_def,herd_well_formed_threads_def] THEN
    met []
  ) THEN
  ps [actions_respect_location_kinds_def,RES_FORALL,FORALL_PROD] THEN
  specpat `!p. P` `a` THEN
  `actions_respect_location_kinds X.acts X.lkh` by ALL_TAC THEN1
  ( ps [herd_provides_def,herd_well_formed_threads_def] THEN
    met []
  ) THEN
  pats `P` [actions_respect_location_kinds_def,RES_FORALL,FORALL_PROD] THEN
  specpat `!p. P` `b` THEN	
  ps [atomicloc_def,nonatomicloc_def,is_at_non_atomic_location_def,is_at_atomic_location_def,is_write_def,is_read_def,loc_of_def] THEN
  Cases_on `a` THEN fs [] THEN
  Cases_on `b` THEN fs [] THEN
  rw [] THEN
  ps [] THEN
  met []
)
  
val paper_rf_then_rf = store_thm("paper_rf_then_rf",
``( !X. (c11_base$hb X = c11_base_intermediate$hb X) /\ herd_provides X /\
      c11_base$Coh X /\ c11_base$Hb X /\
      c11_base$Rf X ==>
      atomicRfConsistent X
  )``,
  rw [] THEN
  `both_coh X` by ps [both_coh_def] THEN
  ps [c11_base_intermediateTheory.atomicRfConsistent_def,
      c11_base_intermediateTheory.hbarf_def,
      c11_base_intermediateTheory.arf_def,
      c11_baseTheory.Rf_def,
     herdTheory.irreflexive_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD] THEN
  rw [] THEN
  ps [seq_def,rcomp_def,FORALL_PROD,stor_def] THEN
  rw [] THEN
  CCONTR_TAC THEN
  ps [] THEN
  specpat `!p. P` `z` THEN
  specpat `!p. P` `p_1` THEN
  ps [atomicloc_def,nonatomicloc_def,is_at_non_atomic_location_def,is_at_atomic_location_def] THEN
  met [rf_then_stuff]
)
				  


val rf_then_paper_rf = store_thm("rf_then_paper_rf",
``( !X. (c11_base$hb X = c11_base_intermediate$hb X) /\ herd_provides X /\
    atomicRfConsistent X /\
    c11_base_intermediate$coherence X /\ c11_base_intermediate$hbAcyclic X /\
    c11_base$NaRf X==>
    c11_base$Rf X
  )``,
  rw [] THEN
  `both_coh X` by ps [both_coh_def] THEN
  ps [c11_base_intermediateTheory.atomicRfConsistent_def,
      c11_base_intermediateTheory.hbarf_def,
      c11_base_intermediateTheory.arf_def,
      c11_baseTheory.Rf_def,
      c11_baseTheory.NaRf_def,
     herdTheory.irreflexive_def,lem_is_irreflexive_def,is_empty_def,RES_FORALL,FORALL_PROD,DIFF_DEF] THEN
  rw [] THEN
  ps [seq_def,rcomp_def,FORALL_PROD,stor_def,EXTENSION,c11_baseTheory.vis_def,c11_baseTheory.hbl_def,compr_def] THEN
  rw [] THEN
  CCONTR_TAC THEN
  ps [] THEN
  specpat `!p. P` `p_1` THEN
  specpat `!p. P` `z` THEN
  specpat `!p. P` `z` THEN
  specpat `!p. P` `p_1` THEN
  ps [hbAcyclic_def,herdTheory.acyclic_def,herdTheory.irreflexive_def,lem_is_irreflexive_def,FORALL_PROD,RES_FORALL,tch_def] THEN
  met [rf_then_loc,tc_cases]
)



				
val paper_rmw_then_rmw = store_thm("paper_rmw_then_rmw",
``( !X. (c11_base$hb X = c11_base_intermediate$hb X) /\ herd_provides X /\
    c11_base_intermediate$coherence X /\ c11_base_intermediate$hbAcyclic X /\
    c11_base_intermediate$rmwAtomicity X ==>
    c11_base$Rmw X
  )``,
  rw [] THEN
  `both_coh X` by ps [both_coh_def] THEN
  ps [c11_base_intermediateTheory.rmwAtomicity_def,
      c11_base_intermediateTheory.rmwv_def,
      c11_baseTheory.Rmw_def,
     herdTheory.irreflexive_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD] THEN
  rw [] THEN
  ps [rc_def,seq_def,rcomp_def,herdTheory.fr_def,hbl_def,inv_def,lem_converse_def,FORALL_PROD,SWAP_def,c11_baseTheory.mo_def,loc_def,stor_def] THEN
  rw [] THEN
  CCONTR_TAC THEN
  ps [] THEN
  rw [] THEN
  met [rf_then_stuff,R_W_rmw]
)


				  



val rmw_then_paper_rmw = store_thm("rmw_then_paper_rmw",
``( !X. (c11_base$hb X = c11_base_intermediate$hb X) /\ herd_provides X /\
    c11_base$Coh X /\ c11_base$Hb X /\
    c11_base$Rmw X ==>
    c11_base_intermediate$rmwAtomicity X
  )``,
  rw [] THEN
  `both_coh X` by ps [both_coh_def] THEN
  ps [c11_base_intermediateTheory.rmwAtomicity_def,
      c11_base_intermediateTheory.rmwv_def,
      c11_baseTheory.Rmw_def,
     herdTheory.irreflexive_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD] THEN
  rw [] THEN
  ps [rc_def,seq_def,rcomp_def,herdTheory.fr_def,hbl_def,inv_def,lem_converse_def,FORALL_PROD,SWAP_def,c11_baseTheory.mo_def,loc_def,stor_def] THEN
  rw [] THEN
  CCONTR_TAC THEN
  ps [] THEN
  rw [] THEN
  met [rf_then_stuff,R_W_rmw]
)

				  
				
val baseProvidesEquivalent = store_thm("baseProvidesEquivalent",
``(  c11_base$provides_clauses = c11_base_intermediate$provides_clauses
  )``,
  ps [EXTENSION,IN_DEF,
      c11_baseTheory.provides_clauses_def,
      c11_base_intermediateTheory.provides_clauses_def] THEN
  rw [] THEN
  EQ_TAC THEN
  ps [] THEN
  rw [] THEN
  `c11_base$hb x = c11_base_intermediate$hb x` by ps [both_coh_def,hbEquivalent] THEN
  `NaRf x` by fs [nonAtomicRfConsistent_def,c11_baseTheory.NaRf_def,narf_def,vse_def,c11_baseTheory.vis_def,c11_baseTheory.hbl_def,c11_base_intermediateTheory.hbl_def,c11_baseTheory.hbl_def] THENL
  [ ps [hbAcyclic_def,c11_baseTheory.Hb_def],
    met [coh_then_paper_coh,paper_coh_then_coh],
    met [paper_rmw_then_rmw,rmw_then_paper_rmw],
    ps [nonAtomicRfConsistent_def,c11_baseTheory.NaRf_def,narf_def,vse_def,c11_baseTheory.vis_def,hbl_def,c11_baseTheory.hbl_def,c11_base_intermediateTheory.hbl_def],
    met[paper_rf_then_rf,rf_then_paper_rf],
    ps [hbAgreesLo_def,c11_baseTheory.Lo1_def],
    ps [consistentLo_def,c11_baseTheory.Lo2_def,mv_def],
    ps [hbAcyclic_def,c11_baseTheory.Hb_def],
    met [coh_then_paper_coh,paper_coh_then_coh],
    met [rf_then_paper_rf],
    met [paper_rmw_then_rmw,rmw_then_paper_rmw],
    ps [hbAgreesLo_def,c11_baseTheory.Lo1_def],
    ps [consistentLo_def,c11_baseTheory.Lo2_def,mv_def]
  ]
)


val provides_then_both_coh = store_thm("provides_then_both_coh",
``( !X. c11_base$provides_clauses X \/ c11_base_intermediate$provides_clauses X ==>
both_coh X
  )``,
  rw [] THEN
  ps [both_coh_def,hbEquivalent,c11_baseTheory.provides_clauses_def,c11_base_intermediateTheory.provides_clauses_def]
)

val sc_subset_acts = store_thm("sc_subset_acts",
``(  !X a. a IN sc X ==> a IN X.acts
  )``,
  ps [sc_def]
)


val both_then_cowwpo = store_thm("both_then_cowwpo",
``(  !X. both_coh X ==>
     ~(?a b. (a,b) IN X.po /\ (b,a) IN X.co)
  )``,
  rw [] THEN
  CCONTR_TAC THEN
  ps [] THEN
  `(a,b) IN hbp X /\ (a,b) IN hbo X` by ps [hbp_def,hbo_def,c11_baseTheory.hb_def,c11_base_intermediateTheory.hb_def,c11_baseTheory.sb_def] THEN
  met [both_then_coww]
)


				
val scEquivalent1 = store_thm("scEquivalent1",
``(  !X a b. both_coh X /\ (c11_base$hb X = c11_base_intermediate$hb X) /\
  (a,b) IN cross (sc X) (sc X) INTER cpp11_paper$scp X DIFF stor (unis X) ==> (a,b) IN cpp11$scp X
  )``,
  rw [] THEN
  ps [cross_def,stor_def,unis_def] THENL
  [ ALL_TAC, ps [sc_def] ] THEN
  ps [cpp11_paperTheory.scp_def,cpp11Theory.scp_def] THENL
  [ DISJ1_TAC,
    (DISJ2_TAC THEN DISJ1_TAC),
    (DISJ2_TAC THEN DISJ2_TAC),
    (DISJ2_TAC THEN DISJ2_TAC),
    (DISJ2_TAC THEN DISJ2_TAC),
    (DISJ2_TAC THEN DISJ2_TAC),
    (DISJ2_TAC THEN DISJ2_TAC)
  ] THEN
  ps [r1_def,r2_def,r3_def,r4_def,r5_def,r6_def,r7_def,schb_def,scco_def,scfr_def] THEN
  ps [c11_baseTheory.sbf_def,c11_baseTheory.fsb_def,unis_def,stor_def,seq_def,rcomp_def,compr_def,c11_base_intermediateTheory.hbl_def,cross_def,rc_def,c11_baseTheory.mo_def,c11_baseTheory.sb_def,inv_def,lem_converse_def,FORALL_PROD,EXISTS_PROD,c11_baseTheory.fr_def,herdTheory.fr_def,SWAP_def,c11_base_intermediateTheory.idf_def,c11_base_intermediateTheory.fid_def,c11_baseTheory.hbl_def,c11_baseTheory.fr_def,idh_def] THEN
  met [sc_subset_acts]
)



val scEquivalent2 = store_thm("scEquivalent2",
``(  !X a b. both_coh X /\ (c11_base$hb X = c11_base_intermediate$hb X) /\
  (a,b) IN cpp11$scp X ==>
  (a,b) IN cross (sc X) (sc X) INTER cpp11_paper$scp X DIFF stor (unis X)
  )``,
  rw [] THENL
  [ ps [cpp11Theory.scp_def,
        cpp11Theory.schb_def,
	cpp11Theory.scco_def,
	cpp11Theory.scfr_def,
        seq_def,rcomp_def,stor_def,cross_def],
    ALL_TAC,
    ( ps [stor_def,unis_def] THEN
      DISJ1_TAC THEN
      ps [cpp11Theory.scp_def,cpp11Theory.schb_def,cpp11Theory.scco_def,cpp11Theory.scfr_def,stor_def,unis_def,compr_def
         ] THENL
      [ ( ps [both_coh_def,c11_baseTheory.Hb_def,c11_base_intermediateTheory.hbAcyclic_def,herdTheory.acyclic_def,herdTheory.irreflexive_def,lem_is_irreflexive_def,RES_FORALL,FORALL_PROD,tch_def,seq_def,rcomp_def] THEN
          met [tc_cases]
	),
	ALL_TAC,
        met [],met [],met [],met [],met []
      ] THEN
      ps [seq_def,rcomp_def,c11_base_intermediateTheory.idf_def,c11_base_intermediateTheory.fid_def,rc_def,stor_def] THEN
      CCONTR_TAC THEN rw [] THEN
      `transitive X.po` by ps [both_coh_def,herd_provides_def,herd_well_formed_threads_def,well_formed_po_def,lem_is_strict_partial_order_def] THEN
      met [both_then_cowwpo,both_coh_co_over_acts,transitive_def]
    )
  ] THEN
  ps [cpp11_paperTheory.scp_def,cpp11Theory.scp_def] THENL
  [ ( DISJ1_TAC THEN
      ps [r1_def,schb_def,stor_def,seq_def,rcomp_def]
    ),
    ( DISJ2_TAC THEN DISJ1_TAC THEN
      ps [r2_def,scco_def,rc_def,c11_baseTheory.mo_def,c11_baseTheory.sb_def,stor_def,seq_def,rcomp_def,c11_base_intermediateTheory.idf_def,c11_base_intermediateTheory.fid_def,c11_baseTheory.sbf_def,c11_baseTheory.fsb_def] THEN
      met []
    ),
    ALL_TAC
  ] THEN
  ps [scfr_def,stor_def,hbl_def,seq_def,rcomp_def,inv_def,lem_converse_def,FORALL_PROD,EXISTS_PROD,SWAP_def,compr_def,unis_def] THEN
  DISJ2_TAC THEN DISJ2_TAC THENL
  [ (DISJ2_TAC THEN DISJ1_TAC),
    (DISJ1_TAC),
    (DISJ2_TAC THEN DISJ2_TAC THEN DISJ1_TAC),
    (DISJ2_TAC THEN DISJ2_TAC THEN DISJ2_TAC THEN DISJ1_TAC),
    (DISJ2_TAC THEN DISJ2_TAC THEN DISJ2_TAC THEN DISJ2_TAC),
    met [], met [], met [], met [], met []
  ] THEN
  ps [r3_def,r4_def,r5_def,r6_def,r7_def] THEN
  ps [c11_baseTheory.sbf_def,c11_baseTheory.fsb_def,stor_def,seq_def,rcomp_def,compr_def,c11_base_intermediateTheory.hbl_def,cross_def,rc_def,c11_baseTheory.mo_def,c11_baseTheory.sb_def,herdTheory.inv_def,lem_converse_def,FORALL_PROD,EXISTS_PROD,herdTheory.fr_def,SWAP_def,c11_baseTheory.hbl_def,c11_baseTheory.fr_def,idh_def] THEN
  met []
)



val scEquivalent = store_thm("scEquivalent",
``(  !X. both_coh X /\ (c11_base$hb X = c11_base_intermediate$hb X) ==>
(cross (sc X) (sc X) INTER cpp11_paper$scp X DIFF stor (unis X) = cpp11$scp X)
  )``,
  rw [] THEN
  psg [EXTENSION,FORALL_PROD] THEN
  rw [] THEN
  EQ_TAC THEN
  rw [] THENL
  [ ( ASSUME_TAC scEquivalent1 THEN
      ps []
    ),
    ALL_TAC,
    ALL_TAC,
    ALL_TAC
  ] THEN
  ASSUME_TAC scEquivalent2 THEN
  ps []
)


  
				      
val providesEquivalent = store_thm("providesEquivalent",
``(  cpp11_paper$provides_clauses = cpp11$provides_clauses
  )``,
  ps [EXTENSION,IN_DEF,
      cpp11_paperTheory.provides_clauses_def,
      cpp11Theory.provides_clauses_def] THEN
  rw [] THEN
  EQ_TAC THEN
  ps [] THEN
  rw [] THEN
  `c11_base$hb x = c11_base_intermediate$hb x` by met [hbEquivalent,provides_then_both_coh] THENL
  [ met [baseProvidesEquivalent],
    ALL_TAC,
    met [baseProvidesEquivalent],
    ALL_TAC
  ] THEN
  ps [cpp11_paperTheory.Spartial_def,cpp11Theory.scConsistent_def] THEN
  met [scEquivalent,provides_then_both_coh]
)

val W_subset_acts = store_thm("W_subset_acts",
``(  !X a. a IN W X ==> a IN X.acts
  )``,
  ps [W_def]
)

				      
val requiresEquivalent = store_thm("requiresEquivalent",
``( !X. cpp11_paper$provides_clauses X \/ cpp11$provides_clauses X ==>
(cpp11_paper$requires_clauses X = cpp11$requires_clauses X)
  )``,
  STRIP_TAC THEN
  `(c11_base$hb X = c11_base_intermediate$hb X) ==> (cpp11_paper$requires_clauses X = cpp11$requires_clauses X)` by ALL_TAC THENL
  [ALL_TAC,met [hbEquivalent,provides_then_both_coh,cpp11_paperTheory.provides_clauses_def,cpp11Theory.provides_clauses_def]] THEN
  rw [] THEN
  ps [c11_baseTheory.requires_clauses_def,
      c11_base_intermediateTheory.requires_clauses_def,
      cpp11_paperTheory.requires_clauses_def,
      cpp11Theory.requires_clauses_def] THEN
  rw [] THEN
  EQ_TAC THEN
  rw [] THENL
  [  ps [dataRace_def,Dr_def],
     ps [cpp11_paperTheory.unsequencedRace_def,c11_base_intermediateTheory.unsequencedRace_def],
     ps [cpp11_paperTheory.badLock_def,c11_base_intermediateTheory.badLock_def],
     ps [cpp11_paperTheory.badUnlock_def,c11_base_intermediateTheory.badUnlock_def],
     ps [dataRace_def,Dr_def],
     ps [cpp11_paperTheory.unsequencedRace_def,c11_base_intermediateTheory.unsequencedRace_def],
     ps [cpp11_paperTheory.badLock_def,c11_base_intermediateTheory.badLock_def],
     ps [cpp11_paperTheory.badUnlock_def,c11_base_intermediateTheory.badUnlock_def]
  ] THEN
  ps [c11_base_intermediateTheory.dr_def,c11_baseTheory.dr_def,c11_baseTheory.cnf_def,unis_def,ext_def,cross_def,compr_def,inv_def,DIFF_DEF,lem_converse_def,is_empty_def,FORALL_PROD,SWAP_def,EXTENSION,loc_def,c11_baseTheory.ur_def,c11_base_intermediateTheory.ur_def,int_def,stor_def,c11_baseTheory.sb_def,c11_baseTheory.bl_def,c11_base_intermediateTheory.bl_def,c11_baseTheory.lu_def,c11_base_intermediateTheory.lu_def,c11_baseTheory.losbwoul_def,c11_base_intermediateTheory.lopowoul_def] THEN
  rw [] THEN
  met [W_subset_acts]
)


				  
val HerdsEquivalent = store_thm("HerdsEquivalent",
``( !opsem p. calculate_program_behaviour cpp11_paper$model opsem p =
              calculate_program_behaviour cpp11$model opsem p
  )``,
  rw [] THEN
  fs [herdTheory.calculate_program_behaviour_def,
      cpp11_paperTheory.model_def,
      cpp11Theory.model_def,
      RES_FORALL,LET_DEF] THEN
  rw [] THEN
  ps [EXTENSION] THEN
  met [requiresEquivalent,providesEquivalent]
)
      

      


     
val lemHerdEquivalent_paper = store_thm("lemHerdEquivalent_paper",
``(! herd_opsem opsem p.
  lemHerdCondition herd_opsem opsem p ==>
  lemHerdBehaviourEquivalent with_consume_memory_model cpp11_paper$model herd_opsem opsem p
)``,
  rw [] THEN
  ps [lemHerdBehaviourEquivalent_def] THEN
  `calculate_program_behaviour cpp11_paper$model herd_opsem p = calculate_program_behaviour cpp11$model herd_opsem p` by met [HerdsEquivalent] THEN
  ASSUME_TAC lemHerdEquivalent THEN
  ps [lemHerdBehaviourEquivalent_def]
)

				 

val _ = export_theory()

